# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
# Embedded Debug Tools

The most important tools:

- GDB: `emdbg.debug.gdb`.
- FMU: `emdbg.bench.fmu`.
- Patches: `emdbg.patch.set`.
"""

from importlib.metadata import version, PackageNotFoundError
try:
    __version__ = version("emdbg")
except PackageNotFoundError:
    __version__ = "0.0.0"


from . import analyze
from . import bench
from . import debug
from . import power
from . import serial
from . import io

from . import logger
from . import utils

# Silence warnings about path import order when calling modules directly
import sys
if not sys.warnoptions:
    import warnings
    warnings.filterwarnings('ignore', category=RuntimeWarning, module='runpy')



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
.. include:: jlink.md
"""

from __future__ import annotations
import os
import time
import signal
import platform
import subprocess
import logging
import tempfile
from pathlib import Path

from .backend import ProbeBackend
from . import gdb

LOGGER = logging.getLogger("debug:jlink")


# -----------------------------------------------------------------------------
class JLinkBackend(ProbeBackend):
    """
    J-Link specific debug backend implementation. Starts the `JLinkGDBServer` in
    its own subprocess and connects GDB to port `2331` of the localhost.

    See `call()` for additional information.
    """
    def __init__(self, device: str, speed: int, rtos: Path = None):
        """
        :param device: part name of the microcontroller to debug.
        :param speed: SWD connection baudrate in kHz.
        :param rtos: Path to RTOS plugin for thread-aware debugging.
        """
        super().__init__(":2331")
        self.device = device
        self.speed = speed
        self.rtos = rtos
        self.process = None
        self.name = "jlink"

    def start(self):
        self.process = call(self.device, self.speed, self.rtos,
                            blocking=False, silent=True)
        LOGGER.info(f"Starting {self.process.pid}...")

    def stop(self):
        if self.process is not None:
            LOGGER.info(f"Stopping {self.process.pid}.")
            os.killpg(os.getpgid(self.process.pid), signal.SIGINT)
            os.waitpid(os.getpgid(self.process.pid), 0)
            self.process = None


# -----------------------------------------------------------------------------
def call(device: str, speed: int, rtos: Path = None, blocking: bool = True,
         silent: bool = False) -> "int | subprocess.Popen":
    """
    Starts the `JLinkGDBServer` and connects to the microcontroller without
    resetting the device and without GUI. You can overwrite the default binary
    by exporting an alternative in your environment:

    ```sh
    export PX4_JLINK_GDB_SERVER=path/to/JLinkGDBServer
    ```

    :param device: part name of the microcontroller to debug.
    :param speed: SWD connection baudrate in kHz.
    :param rtos: Path to RTOS plugin for thread-aware debugging.
    :param blocking: Run in current process as a blocking call.
                     Set to `False` to run in a new subprocess.
    :param silent: Disable any reporting from `JLinkGDBServer`.

    :return: The process return code if `blocking` or the Popen object.
    """
    binary = os.environ.get("PX4_JLINK_GDB_SERVER", "JLinkGDBServer")

    command_jlink = f"{binary} -device {device} -speed {speed} -if swd -noreset -nogui"
    if rtos: command_jlink += f" -rtos {Path(rtos).absolute()}"
    if silent: command_jlink += " -silent"

    LOGGER.debug(command_jlink)

    kwargs = {"cwd": os.getcwd(), "shell": True}
    if blocking:
        return subprocess.call(command_jlink, **kwargs)

    # Disconnect the JLink process from this process, so that any Ctrl-C usage
    # in GDB does not get passed on to the JLink process.
    kwargs["start_new_session"] = True
    return subprocess.Popen(command_jlink, **kwargs)


# -----------------------------------------------------------------------------
def itm(device: str, baudrate: int = None, port: int = None) -> int:
    """
    Launches `JLinkSWOViewer`, connects to the microcontroller. You can
    overwrite the default binary by exporting an alternative in your environment:

    ```sh
    export PX4_JLINK_SWO_VIEWER=path/to/JLinkSWOViewer
    ```

    :param device: part name of the microcontroller to debug.
    :param baudrate: optional frequency of the SWO connection.
    :param port: ITM port to display at startup. You can modify this at runtime.

    :return: the process return code
    """
    binary = os.environ.get("PX4_JLINK_SWO_VIEWER", "JLinkSWOViewer")
    command_jlink = f"{binary} -device {device} -itmport {port or 0}"
    if baudrate: command_jlink += f" -swofreq {baudrate}"
    try:
        return subprocess.call(command_jlink, shell=True)
    except KeyboardInterrupt:
        pass


def rtt(backend: JLinkBackend, channel: int = 0) -> int:
    """
    Launches the backend in the background and connects a telnet client to the
    Real-Time Transfer process. You can disconnect with Ctrl+D.

    :param backend: A J-Link backend object.
    :param channel: The RTT channel to connect to.

    :return: the process return code
    """
    import telnetlib
    # Start JLinkGDBServer in the background
    with backend.scope():
        with telnetlib.Telnet("localhost", 19021) as tn:
            try:
                tn.interact()
            except KeyboardInterrupt:
                pass
    return 0

# -----------------------------------------------------------------------------
def program(device: str, speed: int, source: Path, load_addr: int = None) -> int:
    """
    Loads the source file into the microcontroller and resets the device. You can
    overwrite the default binary by exporting an alternative in your environment:

    ```sh
    export PX4_JLINK=path/to/JLinkExe
    ```

    :param device: part name of the microcontroller to program.
    :param speed: SWD connection baudrate in kHz.
    :param source: path to a `.bin`, `.hex`, or `.elf` file to upload.
    :param load_addr: Specifies the start address to load a `.bin` file.

    :return: the process return code of JLink
    """
    binary = os.environ.get("PX4_JLINK", "JLinkExe")
    if load_addr is None and Path(source).suffix == ".bin":
        load_addr = 0x0800_0000
    with tempfile.NamedTemporaryFile() as fcmd:
        # LoadFile erases the correct amount of sectors, then writes the new
        # binary or hex file. We still need to reset afterwards.
        Path(fcmd.name).write_text(f"LoadFile {source} {load_addr or ''}\nReset\nExit")
        jcmd = f"{binary} -device {device} -if swd -speed {speed} " \
               f"-autoconnect 1 -nogui 1 -commandfile {fcmd.name}"
        LOGGER.debug(jcmd)
        return subprocess.call(jcmd, shell=True)

def reset(device: str, speed: int) -> int:
    """
    Resets the device via JLink.

    :param device: part name of the microcontroller to program. You can
                   overwrite the default binary by exporting an alternative in
                   your environment:
                   ```sh
                   export PX4_JLINK=path/to/JLinkExe
                   ```
    :param speed: SWD connection baudrate in kHz.

    :return: the process return code of JLink
    """
    binary = os.environ.get("PX4_JLINK", "JLinkExe")
    with tempfile.NamedTemporaryFile() as fcmd:
        Path(fcmd.name).write_text(f"Reset\nExit")
        jcmd = f"{binary} -device {device} -if swd -speed {speed} " \
               f"-autoconnect 1 -nogui 1 -commandfile {fcmd.name}"
        LOGGER.debug(jcmd)
        return subprocess.call(jcmd, shell=True)


# -----------------------------------------------------------------------------
def _add_subparser(subparser):
    parser = subparser.add_parser("jlink", help="Use JLink as Backend.")
    parser.add_argument(
            "-device",
            required=True,
            help="Connect to this device.")
    parser.add_argument(
            "-speed",
            type=int,
            default=12000,
            help="SWO baudrate in kHz.")
    parser.set_defaults(backend=lambda args: JLinkBackend(args.device))
    return parser


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    import argparse
    import emdbg

    parser = argparse.ArgumentParser(
        description="JLink debug probe: Upload, reset, and logging")
    parser.add_argument(
        "-device",
        required=True,
        help="Connect to this device.")
    parser.add_argument(
        "-speed",
        type=int,
        default=12000,
        help="SWD baudrate in kHz.")
    parser.add_argument(
        "-v",
        dest="verbosity",
        action="count",
        help="Verbosity level.")

    subparsers = parser.add_subparsers(title="Command", dest="command")

    subparsers.add_parser("reset", help="Reset the device.")

    subparsers.add_parser("run", help="Run JLinkGDBServer.")

    upload_parser = subparsers.add_parser("upload", help="Upload firmware.")
    upload_parser.add_argument(
        "--source",
        required=True,
        help="The firmware to upload: `.bin`, `.hex`, or `.elf` file")
    upload_parser.add_argument(
        "--load-addr",
        help="The load address of the binary file")

    rtt_parser = subparsers.add_parser("rtt", help="Connect to an RTT channel.")
    rtt_parser.add_argument(
        "--channel",
        type=int,
        default=0,
        help="The RTT channel to connect to.")

    itm_parser = subparsers.add_parser("itm", help="Show ITM log output.")
    itm_parser.add_argument(
        "--baudrate",
        type=int,
        help="Force a baudrate instead of auto-selecting it.")
    itm_parser.add_argument(
        "--channel",
        type=int,
        help="The channel to output.")

    args = parser.parse_args()
    emdbg.logger.configure(args.verbosity)

    if args.command == "reset":
        exit(reset(args.device, args.speed))

    if args.command == "run":
        exit(call(args.device, args.speed, blocking=True))

    if args.command == "upload":
        exit(program(args.device, args.speed, os.path.abspath(args.source)))

    if args.command == "rtt":
        backend = JLinkBackend(args.device, args.speed)
        exit(rtt(backend, args.channel))

    if args.command == "itm":
        exit(itm(args.device, args.baudrate, args.channel))

    LOGGER.error("Unknown command!")
    exit(1)



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
from functools import cached_property
from dataclasses import dataclass
from . import utils
from .base import Base


class Semaphore(Base):
    """
    Accessor for a NuttX semaphore
    """

    @dataclass
    class Holder:
        task: "emdbg.debug.px4.task.Task"
        """The task this holder belongs to"""
        count: int
        """How many times the task is holding this semaphore"""

    def __init__(self, gdb, sem_ptr: "gdb.Value"):
        super().__init__(gdb)
        self._sem = sem_ptr
        self.count: int = self._sem["semcount"]
        """Semaphore count, positive if available, 0 if taken, negative if tasks are waiting"""

    @cached_property
    def flags(self) -> int:
        """:return: The flags field of the semaphore"""
        return utils.gdb_getfield(self._sem, "flags", 0b1)

    @cached_property
    def holders(self) -> list[Holder]:
        """:return: a list of semaphore holders"""
        from .task import Task

        #if CONFIG_SEM_PREALLOCHOLDERS == 0
        if holder := utils.gdb_getfield(self._sem, "holder"):
            return [self.Holder(Task(self._gdb, holder[0]["htcb"]), int(holder[0]["counts"]))]

        #if CONFIG_SEM_PREALLOCHOLDERS > 0
        holders = []
        holder = self._sem["hhead"]
        while(holder):
            holders.append(self.Holder(Task(self._gdb, holder["htcb"]), int(holder["counts"])))
            holder = holder["flink"]
        return holders

    @property
    def has_priority_inheritance(self) -> bool:
        """:return: `True` if the semaphore supports priority inheritance"""
        return not (self.flags & 0b1)

    def to_string(self) -> str:
        sc = self.count
        state = "taken" if sc <= 0 else "available"
        ostr = f"{state} ({sc}{'i' if self.has_priority_inheritance else ''})"
        if hstr := [f"{h.count}x {h.task.name}" for h in self.holders]:
            ostr += f" holders: {', '.join(hstr)}"
        return ostr



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

import logging
import logging.config

VERBOSITY = 0
LEVEL = logging.WARNING


def configure(verbosity: int = None):
    """
    Configures the logging of the entire module.

    :param verbosity:
        Set the logging level: ≥-1: ERROR, ≥0: WARNING, ≥1: INFO, ≥2: DEBUG.
        Default is None=0=WARNING.
    """
    if verbosity is None:
        verbosity = 0

    global VERBOSITY
    VERBOSITY = verbosity

    global LEVEL
    if verbosity >= 2:
        LEVEL = logging.DEBUG
    elif verbosity == 1:
        LEVEL = logging.INFO
    elif verbosity == 0:
        LEVEL = logging.WARNING
    else:
        LEVEL = logging.ERROR

    # TODO: Custom logger configuration for only the emdbg module
    logging.basicConfig(level=LEVEL)

    # Disable some particularly chatty modules
    logging.getLogger("pygdbmi.IoManager").setLevel(logging.INFO)
    logging.getLogger("graphviz._tools").setLevel(logging.INFO)



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

import gdb, argparse, shlex, re, traceback, functools
from collections import defaultdict
from rich.console import Console
# The import is relative only to the emdbg/debug folder, so that we do not pull
# in any other dependencies
import px4

_CONSOLE = Console(force_terminal=True)

def report_exception(f):
    @functools.wraps(f)
    def inner(*args, **kwargs):
        try: return f(*args, **kwargs)
        except: print(traceback.format_exc())
    return inner


# Wrap functionality into user commands
class PX4_Discover(gdb.Command):
    """
    Find out what device we are connected to.
    """
    def __init__(self):
        super().__init__("px4_discover", gdb.COMMAND_USER)

    @report_exception
    def invoke(self, argument, from_tty):
        _CONSOLE.print(px4.discover_device(gdb))


class PX4_Tasks(gdb.Command):
    """
    Print a table of all NuttX tasks and their current state.
    """
    def __init__(self):
        super().__init__("px4_tasks", gdb.COMMAND_USER)
        self.parser = argparse.ArgumentParser(self.__doc__)
        self.parser.add_argument("-f", "--files", default=False, action="store_true",
                                 help="Print the names of the open files.")

    @report_exception
    def invoke(self, argument, from_tty):
        args = self.parser.parse_args(shlex.split(argument))
        table, output = px4.all_tasks_as_table(gdb, with_file_names=args.files)
        if table is not None:
            _CONSOLE.print(table)
            print(output)
        else:
            print("No tasks found!")


class PX4_Files(gdb.Command):
    """
    Print a table of all open files and who is holding them.
    """
    def __init__(self):
        super().__init__("px4_files", gdb.COMMAND_USER)

    @report_exception
    def invoke(self, argument, from_tty):
        table = px4.all_files_as_table(gdb)
        if table is not None:
            _CONSOLE.print(table)
        else:
            print("No tasks found!")


class PX4_Registers(gdb.Command):
    """
    Print a table of all Cortex-M registers.
    Optional argument is number of columns.
    """
    def __init__(self):
        super().__init__("px4_registers", gdb.COMMAND_USER)

    @report_exception
    def invoke(self, argument, from_tty):
        _CONSOLE.print(px4.all_registers_as_table(gdb, int(argument or 3)))


class PX4_Interrupts(gdb.Command):
    """
    Print a table of all registered NuttX interrupts.
    EPA = Enabled/Pending/Active, P = (Shifted) Priority
    Optional argument is number of columns.
    """
    def __init__(self):
        super().__init__("px4_interrupts", gdb.COMMAND_USER)

    @report_exception
    def invoke(self, argument, from_tty):
        _CONSOLE.print(px4.vector_table_as_table(gdb, int(argument or 1)))


class PX4_Gpios(gdb.Command):
    """
    Print a table of all GPIOs, their configuration and their FMU specific names.
    You can sort the table with the `-s COLUMN` option and filter it by pin name

    """
    def __init__(self):
        super().__init__("px4_gpios", gdb.COMMAND_USER)
        self.header = ["pin", "config", "in", "out", "af", "name", "function"]
        self.parser = argparse.ArgumentParser(self.__doc__)
        self.parser.add_argument("-f", "--filter", help="Regex filter for FMU names.")
        self.parser.add_argument("-ff", "--function-filter", help="Regex filter for FMU functions.")
        self.parser.add_argument("-pf", "--pin-filter", help="Regex filter for GPIO pin names.")
        self.parser.add_argument("-s", "--sort", help="Column name to sort the table by.",
                                 choices=self.header)
        self.parser.add_argument("-c", "--columns", type=int, default=2,
                                 help="Number of columns to print.")

    @report_exception
    def invoke(self, argument, from_tty):
        args = self.parser.parse_args(shlex.split(argument))
        pinout = px4.pinout(gdb, "fmu")
        columns = args.columns
        if args.pin_filter or args.filter or args.function_filter:
            columns = 1
        def pin_filter(row):
            if args.pin_filter is not None and not re.search(args.pin_filter, row[0]):
                return False
            if len(row) > 5:
                if args.filter is not None and not re.search(args.filter, row[5]):
                    return False
                if args.function_filter is not None and not re.search(args.function_filter, row[6]):
                    return False
            return True
        sort_by = self.header.index(args.sort) if args.sort is not None else None
        _CONSOLE.print(px4.all_gpios_as_table(gdb, pinout, pin_filter, sort_by, columns))


class PX4_Backtrace(gdb.Command):
    """
    Print a backtrace of the current frame.
    This works like `backtrace`, except it doesn't fail internal GDB assertions.
    """
    def __init__(self):
        super().__init__("px4_backtrace", gdb.COMMAND_USER)

    @report_exception
    def invoke(self, argument, from_tty):
        print(px4.backtrace(gdb))


class PX4_Switch_Task(gdb.Command):
    """
    Switch to a task PID to inspect the task.
    """
    def __init__(self):
        super().__init__("px4_switch_task", gdb.COMMAND_USER)

    @report_exception
    def invoke(self, argument, from_tty):
        pid = int(argument, 0) if argument else -1
        px4.task_switch(gdb, pid)


class PX4_Relative_Breakpoint(gdb.Command):
    """
    Finds the absolute location of a relative line number offset and then sets
    a breakpoint on that location. Backslashes in the regex pattern are preserved.

    Location format: `file:function:+offset` or `file:function:regex`.
    """
    def __init__(self):
        super().__init__("px4_rbreak", gdb.COMMAND_USER)

    @report_exception
    def invoke(self, argument, from_tty):
        location = px4.utils.gdb_relative_location(gdb, argument)
        gdb.execute(f"break {location}")


class PX4_Coredump(gdb.Command):
    """
    Dump the volatile memories and registers.
    Optional argument is the filename.
    """
    def __init__(self):
        super().__init__("px4_coredump", gdb.COMMAND_USER)
        self.parser = argparse.ArgumentParser(self.__doc__)
        self.parser.add_argument("--memory", action="append",
                                 help="Memory range in `start:size` format.")
        self.parser.add_argument("--flash", action="store_true", default=False,
                                 help="Also dump the non-volatile memory.")
        self.parser.add_argument("--file",
                                 help="Coredump filename, defaults to `coredump_{datetime}.txt`.")

    @report_exception
    def invoke(self, argument, from_tty):
        args = self.parser.parse_args(shlex.split(argument))
        memories = None
        if args.memory:
            memories = [[int(h, 0) for h in m.split(":")] for m in args.memory]
        px4.coredump(gdb, memories, args.flash, args.file)


class PX4_Watch_Peripheral(gdb.Command):
    """
    Visualize the differences in peripheral registers on every GDB stop event.
    """
    def __init__(self, filename):
        super().__init__("px4_pwatch", gdb.COMMAND_USER)
        self.parser = argparse.ArgumentParser(self.__doc__)
        self.parser.add_argument("name", nargs="*",
                                 help="One or more peripheral or peripheral register names: PER or PER.REG .")
        self.parser.add_argument("--add", "-a", action="store_true", default=False,
                                 help="Add these peripherals.")
        self.parser.add_argument("--remove", "-r", action="store_true", default=False,
                                 help="Remove these peripherals.")
        self.parser.add_argument("--reset", "-R", action="store_true", default=False,
                                 help="Reset watcher to peripheral reset values.")
        self.parser.add_argument("--quiet", "-q", action="store_true", default=False,
                                 help="Stop automatically reporting.")
        self.parser.add_argument("--loud", "-l", action="store_true", default=False,
                                 help="Automatically report on GDB stop event.")
        self.parser.add_argument("--all", "-x", action="store_true", default=False,
                                 help="Show all logged changes.")
        self.parser.add_argument("--watch-write", "-ww", action="store_true", default=False,
                                 help="Add a write watchpoint on registers.")
        self.parser.add_argument("--watch-read", "-wr", action="store_true", default=False,
                                 help="Add a read watchpoint on registers.")
        self.do_report = True
        self.last_report = {}
        self.all_report = {}
        self.watchpoints = {}
        self.svd = px4.PeripheralWatcher(gdb, filename)
        gdb.events.stop.connect(self.on_stop)

    @report_exception
    def invoke(self, argument, from_tty):
        args = self.parser.parse_args(shlex.split(argument))
        if args.add:
            for name in args.name:
                report = self.svd.watch(name)
                if args.loud: print(report)
                arange = self.svd.address(name).values()
                amin, amax = min(a[0] for a in arange), max(a[1] for a in arange)
                match (args.watch_read, args.watch_write):
                    case (False, False): command = None
                    case (True,  False): command = "rwatch"
                    case (False, True):  command = "watch"
                    case (True,  True):  command = "awatch"
                if command:
                    ceil2 = 1 << (amax - amin - 1).bit_length()
                    command = f"{command} *(uint8_t[{ceil2}]*){hex(amin)}"
                    print(command)
                    output = gdb.execute(command, to_string=True)
                    print(output)
                    if match := re.match(r"Hardware watchpoint (\d+):", output):
                        self.watchpoints[name] = int(match.group(1))
        elif args.remove:
            for name in args.name:
                self.svd.unwatch(name)
                if name in list(self.watchpoints.keys()):
                    gdb.execute(f"delete {self.watchpoints.pop(name)}")
            if not args.name:
                self.svd.unwatch()
                for name in list(self.watchpoints.keys()):
                    gdb.execute(f"delete {self.watchpoints.pop(name)}")
        elif args.reset:
            for name in args.name:
                self.svd.reset(name)
            if not args.name:
                self.svd.reset()
        elif args.quiet:
            self.do_report = False
        elif args.loud:
            self.do_report = True
        else:
            for name in (args.name or [None]):
                print(self.report(name, args.all))

    @report_exception
    def report(self, name=None, show_all=False):
        report_map = self.all_report if show_all else self.last_report
        output = []
        if name is not None:
            for register in sorted(self.svd._find(name), key=lambda r: r[1].address_offset):
                if report := report_map.get(register, ""):
                    output.append(report)
        else:
            peripherals = defaultdict(list)
            for register in report_map:
                peripherals[register[0]].append(register)
            for peripheral in sorted(peripherals, key=lambda p: p.base_address):
                if report := self.report(peripherals[peripheral], show_all):
                    output.append(f"Differences for {peripheral.name}:")
                    output.append(report)
        return "\n".join(output)

    @report_exception
    def on_stop(self, event):
        self.last_report = {}
        for reg in self.svd._watched:
            if report := self.svd.report(reg):
                self.last_report[reg] = report
                self.all_report[reg] = report
        if self.do_report:
            print(self.report())
        self.svd.update()


class PX4_Show_Peripheral(gdb.Command):
    """
    Show the value and descriptions of one peripherals and optional register.
    Note: This will read register with side-effects!
    """
    def __init__(self, filename):
        super().__init__("px4_pshow", gdb.COMMAND_USER)
        if filename is None:
            filename = px4.device.Device(gdb)._SVD_FILE
        gdb.execute(f"arm loadfile st {filename}")

    @report_exception
    def invoke(self, argument, from_tty):
        gdb.execute(f"arm inspect /hab st {argument}")


# Instantiate all user commands
PX4_Discover()
PX4_Tasks()
PX4_Files()
PX4_Registers()
PX4_Interrupts()
PX4_Gpios()
PX4_Switch_Task()
PX4_Relative_Breakpoint()
PX4_Backtrace()
PX4_Coredump()
PX4_Watch_Peripheral(px4._SVD_FILE)
PX4_Show_Peripheral(px4._SVD_FILE)


# Functions for use in GDB scripts
class PX4_Relative_Location(gdb.Function):
    """
    Finds the absolute location of a relative line number offset.
    Note that backslashes in the regex must be double escaped!!!
    """
    def __init__(self):
        super().__init__("px4_rloc")

    @report_exception
    def invoke(self, location):
        return px4.utils.gdb_relative_location(gdb, location.string())


class PX4_IsValid(gdb.Function):
    """Is a variable valid? = not <optimized out> and not <unavailable>."""
    def __init__(self):
        super().__init__("px4_valid")

    def invoke(self, var):
        if var.is_optimized_out:
            return 0
        descr = str(var)
        if descr in ["<optimized out>", "<unavailable>"]:
            return 0
        else:
            return 1


# Instantiate all user functions
PX4_Relative_Location()
PX4_IsValid()


# Internal development helpers
class PX4_Reload(gdb.Command):
    """
    Reloads the px4 module internally. This can have unwanted side-effects!
    """
    def __init__(self):
        super().__init__("px4_reload", gdb.COMMAND_USER)

    @report_exception
    def invoke(self, argument, from_tty):
        px4.utils._Singleton._instances = {}
        import importlib
        # importlib.reload(px4)
        importlib.reload(px4.base)
        importlib.reload(px4.device)
        importlib.reload(px4.data)
        importlib.reload(px4.semaphore)
        importlib.reload(px4.buffer)
        importlib.reload(px4.svd)
        importlib.reload(px4.system_load)
        importlib.reload(px4.task)
        importlib.reload(px4.utils)
        gdb.execute(f"source /Users/niklaut/dev/Better-Tooling/embedded-debug-tools/src/emdbg/debug/remote/px4.py")

# Instantiate all internal commands
PX4_Reload()

# Pretty Printers
def _px4_print_semaphore(val):
    stype = str(val.type)
    if stype == "sem_t" or stype == "struct sem_s":
        return px4.Semaphore(gdb, val)
    if stype == "struct uart_buffer_s":
        return px4.UartBuffer(gdb, val)
    return None
gdb.pretty_printers.append(_px4_print_semaphore)



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

import time
import logging
LOGGER = logging.getLogger("power")

class Base:
    """
    Interface to enable and disable a single power source.
    """
    def _on(self):
        raise NotImplementedError

    def _off(self):
        raise NotImplementedError

    def on(self, delay: float = 1):
        """Switch relay channel on and wait for `delay` seconds."""
        LOGGER.debug(f"On and wait {delay:.1f}s")
        self._on()
        time.sleep(delay)

    def off(self, delay: float = 1):
        """Switch relay channel off and wait for `delay` seconds."""
        LOGGER.debug(f"Off and wait {delay:.1f}s")
        self._off()
        time.sleep(delay)

    def cycle(self, delay_off: float = 1, delay_on: float = 1):
        """
        Switch relay channel off and wait for `delay_off` seconds, then
        switch relay channel on and wait for `delay_on` seconds.
        """
        self.off(delay_off)
        self.on(delay_on)



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
.. include:: skynode.md
"""

from __future__ import annotations
import time
from pathlib import Path
from .fmu import debug as fmu_debug, Fmu, _arguments

import emdbg

class Skynode(Fmu):
    """
    Skynode Debug Bench with Logic Analyzer and Power relay
    """
    def __init__(self, elf: Path,
                 gdb: "emdbg.debug.remote.gdb.Interface",
                 nsh: "emdbg.serial.protocol.Nsh" = None,
                 power: "emdbg.power.base.Base" = None,
                 io: "emdbg.io.digilent.Digilent" = None):
        super().__init__(elf, gdb, nsh, power, io)

        if self.io is not None:
            self.sdio_cmd = self.io.gpio(0)
            self.sdio_clk = self.io.gpio(1)

    def disturb_sdcard_cmd_line(self, count: int = 15, delay: float = 0.1):
        """
        Toggle the CMD line of the SDIO interface via the Digilent scope to
        trigger an SDCard access failure.
        """
        self.sdio_cmd.set_output(self.sdio_cmd.Config.OpenDrain, self.sdio_cmd.Level.Low)
        time.sleep(1)
        for _ in range(count):
            self.sdio_cmd.high()
            time.sleep(delay)
            self.sdio_cmd.low()
            time.sleep(delay)
        self.sdio_cmd.set_input()

    def disturb_sdcard_clk_line(self, count: int = 5, delay: float = 0.1):
        """
        Toggle the CLK line of the SDIO interface via the Digilent scope to
        trigger an SDCard access failure.
        """
        self.sdio_clk.set_output(self.sdio_clk.Config.PushPull)
        for _ in range(count):
            self.sdio_clk.low()
            time.sleep(delay)
            self.sdio_clk.high()
            time.sleep(delay)
        self.sdio_clk.set_input()


# -----------------------------------------------------------------------------
def debug(px4_directory: Path, target: Path, serial: str = None,
          digilent: str = None, ui: str = None, commands: list[str] = None,
          with_rpyc: bool = False, keep_power_on: bool = True,
          upload: bool = True, backend: str = None) -> Skynode:
    """
    Configures `px4_debug.bench.fmu.bench()` with a YRelay as power connection.
    """
    power = emdbg.power.yocto_relay(channel=1, inverted=True)
    return fmu_debug(px4_directory, target, serial, digilent, power, ui,
                     commands, with_rpyc, keep_power_on, upload, backend,
                     _FmuClass=Skynode)


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    def _modifier(parser):
        parser.add_argument(
            "--turn-off",
            default=False,
            action="store_true",
            help="Turn Skynode off via relay after debugging.")
    args, backend = _arguments("Debug Skynode FMU", _modifier)

    with debug(args.px4_dir, args.target, ui=args.ui, commands=args.commands,
               backend=backend, keep_power_on=not args.turn_off) as gdb_call:
        exit(gdb_call)



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
import re
from . import utils
from functools import cached_property

class Base:
    """
    This base class provides basic abstractions to simplify usage of the GDB
    Python API, which can be a little verbose.

    It also provides a mechanism to invalidate cached properties whenever GDB
    stopped. This allows the use of `@cached_property` when the target is halted
    to cache expensive operations.
    """
    def __init__(self, gdb):
        self._gdb = gdb
        self._inf = gdb.selected_inferior()
        self._arch = self._inf.architecture()
        self.register_names = [r.name for r in self._arch.registers()]
        # Registering a callback for every obj makes GDB *really* slow :(
        global _CACHED_OBJECTS
        if _CACHED_OBJECTS is None:
            gdb.events.stop.connect(_invalidate_cached_properties)
            _CACHED_OBJECTS = []
        _CACHED_OBJECTS.append(self)

    def _invalidate(self):
        for key, value in self.__class__.__dict__.items():
            if isinstance(value, cached_property):
                self.__dict__.pop(key, None)

    @cached_property
    def registers(self) -> dict[str, int]:
        """All register names and unsigned values in the selected frame."""
        return {r: self.read_register(r) for r in self.register_names}

    def read_register(self, name: str) -> int:
        """:return: unsigned value of the named register in the selected frame"""
        frame = self._gdb.selected_frame()
        # This is convoluted because we need to get the raw FPU register values!
        # Otherwise they get cast from double to int, which is wrong.
        value = int(frame.read_register(name).format_string(format="x"), 16)
        return value

    def write_register(self, name: str, value: int):
        """Writes a value into the named register"""
        try:
            # casting to (void*) allows setting the FPU registers with raw!
            self._gdb.execute(f"set ${name} = (void*){int(value):#x}")
        except Exception as e:
            print(e)

    def write_registers(self, values: dict[str, int]):
        """Writes all named registers into the CPU"""
        for name, value in values.items():
            #
            if name in ["control", "faultmask", "primask"]: continue
            # GDB does not know SP, only MSP and PSP
            if name in ["sp", "r13"]:
                name = "msp"
            if name == "msp":
                self.write_register("r13", value)
            # Remove double FP registers
            if name.startswith("d"): continue
            self.write_register(name, value)

    def lookup_static_symbol_in_function(self, symbol_name: str, function_name: str) -> "gdb.Symbol | None":
        """
        Lookup a static symbol inside a function. GDB makes this complicated
        since static symbols inside functions are not accessible directly.

        :return: the symbol if found or `None`
        """
        if (function := self._gdb.lookup_global_symbol(function_name)) is None:
            return None
        function = function.value()
        function_block = self._gdb.block_for_pc(int(function.address))
        for symbol in function_block:
            if symbol.addr_class == self._gdb.SYMBOL_LOC_STATIC:
                if symbol.name == symbol_name:
                    return symbol
        return None

    def lookup_static_symbol_ptr(self, name: str) -> "gdb.Value":
        """:return: a Value to a static symbol name"""
        if symbol := self._gdb.lookup_static_symbol(name):
            return self.value_ptr(symbol)
        return None

    def lookup_global_symbol_ptr(self, name) -> "gdb.Value":
        """:return: a Value to a global symbol name"""
        if symbol := self._gdb.lookup_global_symbol(name):
            return self.value_ptr(symbol)
        return None

    def lookup_static_symbol_in_function_ptr(self, symbol_name: str, function_name: str) -> "gdb.Value | None":
        """:return: a Value to a global symbol name"""
        if symbol := self.lookup_static_symbol_in_function(symbol_name, function_name):
            return self.value_ptr(symbol)
        return None

    def value_ptr(self, symbol: "gdb.Symbol") -> "gdb.Value":
        """
        Convert a symbol into a value. This can be useful if you want to keep a
        "pointer" of the symbol, whose content is always up-to-date, rather than
        a local copy, which will never be updated again.
        """
        return self._gdb.Value(symbol.value().address).cast(symbol.type.pointer())

    def addr_ptr(self, addr: int, type: str) -> "gdb.Value":
        """Cast a memory address to a custom type."""
        return self._gdb.Value(addr).cast(self._gdb.lookup_type(type).pointer())

    def read_memory(self, address: int, size: int) -> memoryview:
        """
        Reads a block of memory and returns its content.
        See [Inferiors](https://sourceware.org/gdb/onlinedocs/gdb/Inferiors-In-Python.html).
        """
        return self._inf.read_memory(address, size)

    def write_memory(self, address: int, buffer, length: int):
        """
        Writes a block of memory to an address.
        See [Inferiors](https://sourceware.org/gdb/onlinedocs/gdb/Inferiors-In-Python.html).
        """
        self._inf.write_memory(address, buffer, length=length)

    def read_uint(self, addr: int, size: int) -> int:
        """Reads an unsigned integer from a memory address"""
        if (itype := {1: "B", 2: "H", 4: "I", 8: "Q"}.get(size)) is None:
            raise ValueError("Unsupported unsigned integer size!")
        return self.read_memory(addr, size).cast(itype)[0]

    def read_int(self, addr: int, size: int) -> int:
        """Reads a signed integer from a memory address"""
        if (itype := {1: "b", 2: "h", 4: "i", 8: "q"}.get(size)) is None:
            raise ValueError("Unsupported signed integer size!")
        return self.read_memory(addr, size).cast(itype)[0]

    def read_string(self, addr: int, encoding: str = None,
                    errors: str = None, length: int = None) -> str:
        """Reads a string of a fixed length, or with 0 termination"""
        kwargs = {"encoding": encoding or "ascii", "errors": errors or "ignore"}
        if length: kwargs["length"] = length
        return self.addr_ptr(addr, "char").string(**kwargs)

    def symtab_line(self, pc: int) -> "gdb.Symtab_and_line":
        """:return: the symbol table and line for a program location"""
        return self._gdb.find_pc_line(int(pc))

    def block(self, pc: int) -> "gdb.Block":
        """:return: the block for a program location"""
        return self._gdb.block_for_pc(int(pc))

    def description_at(self, addr: int) -> str | None:
        """:return: the human-readable symbol description at an address"""
        output = self._gdb.execute(f"info symbol *{int(addr)}", to_string=True)
        if match := re.search(r"(.*?) in section (.*?)", output):
            return match.group(1)
        return None

    def integer_type(self, size: int, signed: bool = True) -> "gdb.Type":
        """:return: The built-in integer type for the size in bits"""
        return self._arch.integer_type(size * 8, signed=True)

    @property
    def uint32(self) -> "gdb.Type":
        """The built-in unsigned 32-bit integer type"""
        return self.integer_type(4, False)

    @property
    def int32(self) -> "gdb.Type":
        """The built-in signed 32-bit integer type"""
        return self.integer_type(4, True)


# Single callback makes GDB much faster, so we do the loop ourselves
_CACHED_OBJECTS = None
def _invalidate_cached_properties(event):
    global _CACHED_OBJECTS
    for obj in _CACHED_OBJECTS:
        if obj: obj._invalidate()



# Copyright (c) 2020-2022, Niklas Hauser
# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
from pathlib import Path
from contextlib import contextmanager


class ProbeBackend:
    """
    Base class for starting and stopping debug probes and attaching GDB to them.
    """
    def __init__(self, remote: str = None):
        """
        :param remote: Extended remote location.
        """
        self.remote = remote
        self.name = "remote"

    def init(self, elf: Path) -> list[str]:
        """
        Returns a list of GDB commands that connect GDB to the debug probe.
        The default implementation returns `target extended-remote {self.remote}`.
        """
        return ["target extended-remote {}".format(self.remote)] if self.remote else []

    def start(self):
        """
        Starts the debug probe as a non-blocking subprocess.
        """
        pass

    def stop(self):
        """
        Halts the debug probe process.
        """
        pass

    @contextmanager
    def scope(self):
        """
        Starts and stops the debug probe process within this scope.
        """
        try:
            self.start()
            yield
        finally:
            self.stop()




# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
from pydwf import DwfLibrary
from pydwf.utilities import openDwfDevice
from contextlib import contextmanager
from functools import lru_cache
import enum
import logging

LOGGER = logging.getLogger("io:dwf")
_DWF = None

@contextmanager
def analog_discovery(identifier: str):
    """
    Starts the DWF library, opens and resets the device and yields it.

    :param identifier: serial number of the device to connect to.
    """
    global _DWF
    if _DWF is None:
        _DWF = DwfLibrary()
    with openDwfDevice(_DWF, serial_number_filter=identifier) as device:
        try:
            LOGGER.info("Starting...")
            device.reset()
            yield Digilent(device)
        finally:
            device.reset()
            LOGGER.debug("Stopping.")

# -----------------------------------------------------------------------------
class DigilentPin:
    """
    Describes a single pin of the Digilent device. Useful only for low-speed
    sampling or driving.
    """
    class Config(enum.Enum):
        """
        Specifies the hardware configuration of the pin.
        """
        Input = enum.auto()
        """Input state"""
        PushPull = enum.auto()
        """Output state with strong high and low driver"""
        OpenDrain = enum.auto()
        """Output state with only a strong low driver"""
        OpenSource = enum.auto()
        """Output state with only a strong high driver"""

    class Level(enum.IntEnum):
        """
        Logic level of the pin, for input and output.
        """
        Low = 0,
        High = 1

    def __init__(self, dio, mask: int):
        """
        :param dio: digitalIO driver of the dwf module.
        :param mask: the single bit mask of the pin.
        """
        self.dio = dio
        self.mask = mask
        self.config = self.Config.Input

    def _set_input(self):
        self.dio.outputEnableSet(self.dio.outputEnableGet() & ~self.mask)

    def _set_output(self):
        self.dio.outputEnableSet(self.dio.outputEnableGet() | self.mask)

    def _set(self):
        self.dio.outputSet(self.dio.outputGet() | self.mask)

    def _reset(self):
        self.dio.outputSet(self.dio.outputGet() & ~self.mask)

    def is_high(self) -> bool:
        """
        .. warning::
            This function does not return the actual pin state, only the set
            output state. Use `read()` for that instead.
        :return: `True` if the pin output state is set high.
        """
        return self.dio.outputGet() & self.mask

    def is_output(self) -> bool:
        """
        :return: `True` if pin is configured as output.
        """
        return self.dio.outputEnableGet() & self.mask

    def set_output(self, config: Config = None, level: Level = None):
        """
        Configures the pin as output with an initial level.
        :param config: pin hardware configuration. Default: `Config.PushPull`
        :param level: initial output level. Default: `Level.High`.
        """
        self.config = self.Config.PushPull if config is None else config
        self.set(level)

    def set(self, level: Level = None):
        """
        Sets an output level. Note that depending on pin configuration, the
        output level may not be achieved.

        :param level: output level. Default: `Level.High`.
        """
        if level is None: level = self.Level.High
        if self.config == self.Config.PushPull:
            if level: self._set()
            else: self._reset()
            self._set_output()
        elif self.config == self.Config.OpenDrain:
            if level:
                self._set_input()
            else:
                self._reset()
                self._set_output()
        elif self.config == self.Config.OpenSource:
            if level:
                self._set()
                self._set_output()
            else:
                self._set_input()

    def high(self):
        """
        Sets the output level to `Level.High`.
        """
        self.set(self.Level.High)

    def low(self):
        """
        Sets the output level to `Level.Low`.
        """
        self.set(self.Level.Low)

    def set_input(self):
        """
        Configures the pin as `Config.Input`.
        """
        self.config = self.Config.Input
        self._set_input()

    def read(self) -> bool:
        """
        :return: the actual input state of the pin.
        """
        return self.dio.inputStatus() & self.mask

# -----------------------------------------------------------------------------
class Digilent:
    """
    Wrapper class of the Digilent device.
    The [DWF Python API][dwf] is already very good and should be used directly
    to configure more advanced use-cases, so this class is very light-weight.
    """
    def __init__(self, device):
        """
        :param device: DWF device
        """
        self.dwf = device

    @lru_cache(100)
    def gpio(self, position: int) -> DigilentPin:
        """
        Creates a pin abstraction.
        :param position: Pin number
        """
        return DigilentPin(self.dwf.digitalIO, 1 << position)

    def _info_digital_io(self):
        info = ["",
            f"Enable Support  = {self.dwf.digitalIO.outputEnableInfo():032b}",
            f"Enabled         = {self.dwf.digitalIO.outputEnableGet():032b}",
            f"Output Support  = {self.dwf.digitalIO.outputInfo():032b}",
            f"Output High/Low = {self.dwf.digitalIO.outputGet():032b}",
            f"Input Support   = {self.dwf.digitalIO.inputInfo():032b}",
            f"Input Read      = {self.dwf.digitalIO.inputStatus():032b}",
        ]
        LOGGER.info("\n".join(info))



# Copyright (c) 2020-2022, Niklas Hauser
# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
.. include:: gdb.md

## Python API
"""

from __future__ import annotations
import os
import sys
import subprocess
import signal
import tempfile
import shlex
import time
import sysconfig
from pathlib import Path
from contextlib import contextmanager

import logging
LOGGER = logging.getLogger("debug:gdb")

from .utils import listify
from .backend import ProbeBackend
from . import remote

# -----------------------------------------------------------------------------
def command_string(backend: ProbeBackend, source: Path = None,
                   config: list[Path] = None, commands: list[str] = None,
                   ui: str = None, svd: Path = None, socket: Path = None,
                   with_python: bool = True, coredump: Path = None) -> str:
    """
    Constructs a command string to launch GDB with the correct options.
    By default, this disables pagination and command confirmation.

    :param backend: a debug backend implementation.
    :param source: Path to a ELF file.
    :param config: List of GDB configuration files.
    :param commands: List of GDB commands to execute during launch.
    :param ui: The frontend configuration for GDB
               - `batch` or None: No UI, launches with `-nx -nh -batch`.
               - `cmd`: Default GDB UI with only a command prompt available.
               - `tui`: Launches in text user interface with layout split.
               - `gdbgui`: Launches in background using GDBGUI as frontend.
    :param svd: Path to the CMSIS-SVD file for the connected device.
    :param socket: Path to socket file, which is used for RPyC communication.
    :param with_python: Uses `arm-none-eabi-gdb-py3` and loads the Python
                        debug modules in `emdbg.debug.px4` as `px4`.
    """
    debug_dir = Path(__file__).parent.resolve()
    cmds = [
        "set pagination off", "set print pretty", "set history save",
        "set mem inaccessible-by-default off", "set confirm off",
        "set filename-display absolute", "set disassemble-next-line on",
        "maintenance set internal-error backtrace on",
        "maintenance set internal-warning backtrace on",
        f"source {debug_dir}/data/orbuculum.gdb",
        f"source {debug_dir}/data/cortex_m.gdb"]
    if (backend_gdb := debug_dir / f"data/{backend.name}.gdb").exists():
        cmds += [f"source {backend_gdb}"]
    cmds += listify(backend.init(source))
    args = [f"-c {coredump}"] if coredump else []
    args += [f'-ex "{a}"' for a in cmds] + ["-q"]
    args += list(map('-x "{}"'.format, listify(config)))
    site_packages = sysconfig.get_paths()["purelib"]

    gdb = "arm-none-eabi-gdb"
    if with_python or socket:
        gdb += "-py3"
        # Import packages from both the host the from emdbg
        args += [f'-ex "python import sys; sys.path.append(\'{site_packages}\');"',
                 f'-ex "python import sys; sys.path.append(\'{debug_dir}\');"']
        # We need to do this terrible hackery since pkg_resources fails on the first import
        args += ['-ex "python exec(\'try: import cmsis_svd;\\\\nexcept: pass\\\\nimport cmsis_svd\')"',
                 '-ex "python exec(\'try: import arm_gdb;\\\\nexcept: pass\\\\nimport arm_gdb\')"']
        # If available, set the default SVD file here
        if svd:
            args += [f'-ex "python import px4; px4._SVD_FILE=\'{svd}\'"']
        # Finally we can import the PX4 GDB user commands
        args += [f'-ex "source {debug_dir}/remote/px4.py"']

    if socket:
        # Import the API bridge that uses rpyc
        args += [f'-ex "python socket_path = \'{socket}\'"',
                 f'-ex "source {debug_dir}/remote/gdb_api_bridge.py"']
        cmd = "{gdb} -nx -nh {args} {source}"

    elif ui is None or "batch" in ui:
        cmd = "{gdb} -nx -nh -batch {args} {source}"

    elif "cmd" in ui:
        cmd = "{gdb} {args} {source}"

    elif "tui" in ui:
        cmd = '{gdb} -tui -ex "layout split" -ex "focus cmd" {args} -ex "refresh" {source}'

    elif "gdbgui" in ui:
        cmd = "gdbgui {source} --gdb-cmd='{gdb} {args} {source}'"

    else:
        raise ValueError("Unknown UI mode! '{}'".format(ui))

    args += list(map('-ex "{}"'.format, listify(commands)))
    return cmd.format(gdb=gdb, args=" ".join(args), source=source or "")


# -----------------------------------------------------------------------------
@contextmanager
def call_rpyc(backend: ProbeBackend, source: Path, config: list[Path] = None,
              commands: list[str] = None, svd: Path = None) -> remote.rpyc.Gdb:
    """
    Launches GDB in the background and connects to it transparently via a RPyC
    bridge. You can therefore use the [GDB Python API][gdbpy] as well as the
    modules in `emdbg.debug.px4` directly from within this process instead of
    using the GDB command line.

    This function returns a `remote.rpyc.Gdb` object, which can be used as a
    substitute for `import gdb`, which is only available *inside* the GDB
    Python environment.
    Note, however, that the access is quite slow, since everything has to be
    communicated through asynchronous IPC.

    [gdbpy]: https://sourceware.org/gdb/onlinedocs/gdb/Python-API.html

    :param backend: a debug backend implementation.
    :param source: Path to a ELF file.
    :param config: List of GDB configuration files.
    :param commands: List of GDB commands to execute during launch.
    :param svd: Path to the CMSIS-SVD file for the connected device.

    :return: `remote.rpyc.Gdb` object which can be used instead of `import gdb`.
    """
    from rpyc import BgServingThread
    from rpyc.utils.factory import unix_connect

    with tempfile.TemporaryDirectory() as socket_dir, backend.scope():
        socket = Path(socket_dir) / "socket"
        gdb_command = command_string(backend, source, config, commands, svd=svd, socket=socket)
        rgdb = None
        try:
            LOGGER.debug(gdb_command)
            rgdb_process = subprocess.Popen(gdb_command, cwd=os.getcwd(),
                                            stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                            start_new_session=True, shell=True)
            LOGGER.info(f"Starting {rgdb_process.pid}...")
            # TODO: Add a timeout so it doesn't get stuck forever
            while(True):
                try:
                    conn = unix_connect(str(socket))
                    break
                except (ConnectionRefusedError, FileNotFoundError):
                    time.sleep(0.1)

            BgServingThread(conn, callback=lambda: None)
            rgdb = remote.rpyc.Gdb(conn, backend, rgdb_process)
            yield rgdb

        except KeyboardInterrupt:
            pass

        finally:
            LOGGER.info(f"Stopping {rgdb_process.pid}.")
            if rgdb is not None:
                rgdb.quit()
            else:
                os.killpg(os.getpgid(rgdb_process.pid), signal.SIGQUIT)
            os.waitpid(os.getpgid(rgdb_process.pid), 0)

# -----------------------------------------------------------------------------
@contextmanager
def call_mi(backend: ProbeBackend, source: Path = None, config: list[Path] = None,
            commands: list[str] = None, svd: Path = None,
            with_python: bool = True) -> remote.mi.Gdb:
    """
    Launches GDB in the background using [pygdbmi][] and connects to its command
    prompt via the [GDB/MI Protocol][gdbmi].
    This method does not allow accessing the Python API directly, instead you
    must issue command strings inside the GDB command prompt.
    However, this method is significantly faster and most stable than `call_rpyc()`.

    :param backend: a debug backend implementation.
    :param source: Path to a ELF file.
    :param config: List of GDB configuration files.
    :param commands: List of GDB commands to execute during launch.
    :param svd: Path to the CMSIS-SVD file for the connected device.
    :param with_python: Uses `arm-none-eabi-gdb-py3` and loads the Python
                        debug modules in `emdbg.debug.px4` as `px4`.

    :return: `remote.mi.Gdb` object which can be used to issue commands and read
             the responses.

    [gdbmi]: https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI.html
    [pygdbmi]: https://cs01.github.io/pygdbmi/
    """
    from pygdbmi.gdbcontroller import GdbController

    gdb_command = command_string(backend, source, config, commands, "cmd", svd, with_python=with_python)
    gdb_command += " --interpreter=mi3"


    with backend.scope():
        rgdb = None
        try:
            LOGGER.info("Starting...")
            LOGGER.debug(gdb_command)
            mi = GdbController(shlex.split(gdb_command))
            rgdb = remote.mi.Gdb(backend, mi)
            yield rgdb

        finally:
            if rgdb is not None:
                rgdb.quit()
            LOGGER.info("Stopping.")


# -----------------------------------------------------------------------------
def _empty_signal_handler(sig, frame):
    pass

def call(backend: ProbeBackend, source: Path = None, config: list[Path] = None,
         commands: list[str] = None, ui: str = None, svd: Path = None,
         with_python: bool = True, coredump: Path = None) -> int:
    """
    Launches the backend in the background and GDB as a blocking process in
    the foreground for user interaction.

    :param backend: a debug backend implementation.
    :param source: Path to a ELF file.
    :param config: List of GDB configuration files.
    :param commands: List of GDB commands to execute during launch.
    :param ui: The frontend configuration for GDB
               - `cmd`: Default GDB UI with only a command prompt available.
               - `tui`: Launches in text user interface with layout split.
               - `gdbgui`: Launches in background using GDBGUI as frontend.
    :param svd: Path to the CMSIS-SVD file for the connected device.
    :param with_python: Uses `arm-none-eabi-gdb-py3` and loads the Python
                        debug modules in `emdbg.debug.px4` as `px4`.
    """
    gdb_command = command_string(backend, source, config, commands, ui, svd,
                                 with_python=with_python, coredump=coredump)

    signal.signal(signal.SIGINT, _empty_signal_handler)
    with backend.scope():
        try:
            LOGGER.info("Starting...")
            LOGGER.debug(gdb_command)
            # This call is now blocking
            return subprocess.call(gdb_command, cwd=os.getcwd(), shell=True)
        except KeyboardInterrupt:
            pass
        finally:
            LOGGER.info("Stopping.")


# -----------------------------------------------------------------------------
def _add_subparser(subparser):
    parser = subparser.add_parser("remote", help="Use a generic extended remote as Backend.")
    parser.add_argument(
        "--port",
        default="localhost:3333",
        help="Connect to this host:port.")
    parser.set_defaults(backend=lambda args: ProbeBackend(args.port))

# -----------------------------------------------------------------------------
if __name__ == "__main__":
    import argparse, emdbg
    from . import jlink, crashdebug, openocd

    parser = argparse.ArgumentParser(description="Debug with GDB")
    parser.add_argument(
        "--elf",
        dest="source",
        type=Path,
        help="The ELF files to use for debugging.")
    parser.add_argument(
        "--ui",
        default="cmd",
        choices=["tui", "gdbgui", "cmd", "batch"],
        help="Use GDB via TUI or GDBGUI.")
    parser.add_argument(
        "-py", "--python",
        dest="with_python",
        action="store_true",
        default=False,
        help="Use GDB with Python API and load PX4 tools.")
    parser.add_argument(
        "-c", "--core",
        type=Path,
        default=None,
        help="Use coredump file.")
    parser.add_argument(
        "--svd",
        type=Path,
        help="The CMSIS-SVD file to use for this device, requires `--python` flag.")
    parser.add_argument(
        "-x",
        dest="config",
        action="append",
        type=Path,
        help="Use these GDB init files.")
    parser.add_argument(
        "-ex",
        dest="commands",
        action="append",
        help="Extra GDB commands.")
    parser.add_argument(
        "-v",
        dest="verbosity",
        action="count",
        default=0,
        help="Verbosity level.")

    subparsers = parser.add_subparsers(title="Backend", dest="backend")

    # Add generic backends
    _add_subparser(subparsers)
    # Add specific backends
    crashdebug._add_subparser(subparsers)
    jlink._add_subparser(subparsers)
    openocd._add_subparser(subparsers)

    args = parser.parse_args()
    emdbg.logger.configure(args.verbosity)

    call(args.backend(args) if args.backend else ProbeBackend(), ui=args.ui,
         source=args.source, config=args.config, commands=args.commands,
         svd=args.svd, with_python=args.with_python, coredump=args.core)




# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
from contextlib import contextmanager
import time


class Interface:
    """
    The interface of all GDB remote access implementations.
    """

    def __init__(self, backend: "emdbg.debug.ProbeBackend"):
        self.backend = backend
        self.interrupt_nesting = 0
        self.interrupted = True
        self.type = "base"

    def execute(self, command: str, timeout: float = 1, to_string: bool = False) -> str | None:
        """
        Executes a command on the GDB command prompt and waits for a response.
        If `to_string` is set, it will return the response as a string.

        :param command: The command string to send to the GDB prompt.
        :param timeout: How long to wait for a response in seconds.
        :param to_string: Capture the response as a string and return it.
        """
        raise NotImplementedError

    @contextmanager
    def interrupt_continue(self):
        """
        Interrupts and yields, then continues.
        If the target was already interrupted before entering this scope, it
        will not continue automatically, so that this context can be used in
        nested calls.
        """
        try:
            interrupted = self.interrupt_and_wait()
            self.interrupt_nesting += 1
            yield
        finally:
            self.interrupt_nesting -= 1
            if self.interrupt_nesting == 0 and interrupted:
                self.continue_nowait()

    def interrupt_and_wait(self) -> bool:
        """
        Interrupt the program and wait until it stops.

        :return: `True` if GDB has been interrupted, `False` if it was already interrupted.
        """
        if not self.interrupted:
            self.execute("interrupt")
            self.interrupted = True
            time.sleep(0.2)
            return True
        return False

    def continue_nowait(self) -> bool:
        """
        Continue the program. Do not wait until it stops again.

        :return: `True` if GDB continues running, `False` if it was already running.
        """
        if self.interrupted:
            self.execute("continue&")
            self.interrupted = False
            time.sleep(0.2)
            return True
        return False

    def quit(self):
        """Terminate GDB."""
        raise NotImplementedError



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
You can generate semaphore boosts logs either at runtime using the
`emdbg.patch.set.semaphore_boostlog` patch, or using the
`px4_calltrace_semaphore_boosts` command of the `emdbg.bench.skynode` module
at debug time.


## Command Line Interface

You can summarize boost logs:

```sh
python3 -m emdbg.analyze.priority log_semaphore_boosts.txt
```

You can also generate a callgraph with boost operations:

```sh
python3 -m emdbg.analyze.callgraph log_semaphore_boosts.txt --svg --type Semaphore
```
"""

from __future__ import annotations
import re
from pathlib import Path
import statistics
import itertools
from collections import defaultdict
from ..debug.px4.utils import format_table
from .utils import read_gdb_log

class _BoostOperation:
    PATTERN = re.compile(r"L(\d+) +(\d+)us> +(0x[0-9a-f]+|<opti-out>): +(.*?) +(\d+) +([\\/_]+) +(\d+) *(.*)")

    def __init__(self, description):
        self.is_valid = False
        if match := self.PATTERN.search(description):
            self.line = int(match.group(1))
            self.uptime = int(match.group(2))
            try:
                self.semaphore = int(match.group(3), 16)
            except:
                self.semaphore = None
            self.task = match.group(4)
            self.prio_from = int(match.group(5))
            self.operation = match.group(6)
            self.prio_to = int(match.group(7))
            self.reason = None if not match.group(8) else match.group(8)
            self.is_valid = True

    def __repr__(self) -> str:
        return f"{self.task} {self.prio_from} {self.operation} {self.prio_to}"


def summarize_semaphore_boostlog(logfile: Path) -> str:
    """
    Analyze a boost log and create a statistical summary and return it.

    :param logfile: The boostlog file to analyze
    :return: A formatted string containing the analysis.
    """
    output = []

    ops = []
    for line in read_gdb_log(logfile).splitlines():
        op = _BoostOperation(line)
        if op.is_valid:
            ops.append(op)
        # else:
        #     print(line)
    if not ops: return "No task boosts found!"
    ops.sort(key=lambda o: o.uptime)


    uptime = ops[-1].uptime
    sample_time = max(o.uptime for o in ops) - min(o.uptime for o in ops)
    output.append(f"Uptime: {uptime/1e6:.1f}s = {uptime/6e7:.1f}min")
    output.append(f"Sample time: {sample_time/1e6:.1f}s = {sample_time/6e7:.1f}min")

    # Print summary statistics
    prios = []
    def _mean(values):
        return int((statistics.fmean(values) if len(values) else 0) / 1e3)
    def _stdev(values):
        if len(values) >= 2: return int(statistics.stdev(values) / 1e3)
        return values[0] if len(values) else 0

    diffs = [n.uptime - p.uptime for p,n in itertools.pairwise(ops)]
    prios.append(["All", len(ops), _mean(diffs), _stdev(diffs)])

    boost_up = [op for op in ops if op.operation == "_/"]
    diffs_up = [n.uptime - p.uptime for p,n in itertools.pairwise(boost_up)]
    prios.append(["Up", len(boost_up), _mean(diffs_up), _stdev(diffs_up)])

    boost_down = [op for op in ops if op.operation == "\\_"]
    diffs_down = [n.uptime - p.uptime for p,n in itertools.pairwise(boost_down)]
    prios.append(["Down", len(boost_down), _mean(diffs_down), _stdev(diffs_down)])

    boost_255 = [op for op in ops if op.operation == "_/" and op.prio_to == 255]
    diffs_255 = [n.uptime - p.uptime for p,n in itertools.pairwise(boost_255)]
    prios.append(["255", len(boost_255), _mean(diffs_255), _stdev(diffs_255)])

    fmtstr = "{:%d}  {:>%d} ~{:>%d}ms ±{:>%d}ms"
    header = ["KIND", "#BOOSTS", "MEAN", "STDEV"]
    output.append("")
    output.append(format_table(fmtstr, header, prios))

    # Print summary statistics per task
    fmtstr = "{:%d}  {:>%d} {:>%d}%%  [{:>%d}, {:>%d}]   {:%d}   {:>%d}  {:%d}"
    header = ["TASK", "#BOOSTS", "PCT", "MIN", "MAX", "REASONS", "#MAX", "REASONS FOR MAX"]
    tasks = defaultdict(list)
    for op in ops:
        tasks[op.task].append(op)
    task_prios = [[
            task,
            boosts := sum(1 for o in ops if o if o.operation == "_/"),
            f"{boosts * 100 / len(boost_up):.1f}",
            min(o.prio_from for o in ops),
            max_prio := max(o.prio_to for o in ops),
            ", ".join(sorted({o.reason for o in ops if o.reason is not None})),
            sum(1 for o in ops if o if o.operation == "_/" and o.prio_to == max_prio),
            ", ".join(sorted({o.reason for o in ops if o.reason is not None and o.prio_to == max_prio}))
        ]
        for task, ops in tasks.items()]
    task_prios.sort(key=lambda t: (t[4], t[3], t[0], t[1]))
    output.append("")
    output.append(format_table(fmtstr, header, task_prios))

    semtasks = defaultdict(list)
    semtasksr = defaultdict(list)
    for op in ops:
        if op.semaphore is not None:
            semtasks[op.semaphore].append(op.task)
            if op.reason: semtasksr[op.semaphore].append(op.reason)
    semtasks = [[hex(s), len(ts), ", ".join(sorted(set(ts))), ", ".join(sorted(set(semtasksr.get(s, []))))] for s, ts in semtasks.items()]
    semtasks.sort(key=lambda s: -s[1])
    fmtstr = "{:%d}  {:>%d}  {:%d}  {:%d}"
    header = ["SEMAPHORE", "#BOOSTS", "BOOSTS THESE TASKS", "BECAUSE OF THESE TASKS"]
    output.append("")
    output.append(format_table(fmtstr, header, semtasks))

    return "\n".join(output)


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Priority Ceiling Analyzer")
    parser.add_argument(
        "file",
        type=Path,
        help="The GDB log containing the semaphore boost trace.")
    args = parser.parse_args()

    print(summarize_semaphore_boostlog(args.file))







# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
from functools import cached_property
from . import utils
from .base import Base
from .device import Device
import logging
LOGGER = logging.getLogger(__name__)


class SystemLoad(Base):
    """
    PX4 system load monitor using the `cpuload.cpp` module inside PX4.
    Do not use this class directly, instead access the cpu load information via
    the `emdbg.debug.px4.task.all_tasks_as_table()` function.
    """
    def __init__(self, gdb):
        super().__init__(gdb)
        self._device = Device(gdb)
        self._system_load = self.lookup_global_symbol_ptr("system_load")
        self._monitor = self.lookup_static_symbol_ptr("cpuload_monitor_all_count")
        self._psl = None
        self.restart()

    def _load(self):
        sl = self._system_load.dereference()
        tasks = {"hrt": self._device.uptime, "start": int(sl["start_time"]), "tasks": {}}
        for task in utils.gdb_iter(sl["tasks"]):
            if not task["valid"]: continue
            tcb = int(task["tcb"])
            total = int(task["total_runtime"])
            tasks["tasks"][tcb] = total
        return tasks

    def restart(self):
        """
        Enables the cpuload monitor in PX4 and loads the first reference value.
        """
        if self._system_load is None: return
        if self._monitor.dereference()["_value"] == 0:
            self._gdb.execute("set cpuload_monitor_all_count = 1")
            self._gdb.execute(f"set system_load.start_time = {self._device.uptime}")
            for ii in range(1, self._system_load["tasks"].type.range()[1]):
                self._gdb.execute(f"set system_load.tasks[{ii}].total_runtime = 0")
                self._gdb.execute(f"set system_load.tasks[{ii}].curr_start_time = 0")
        self._psl = self._load()


    def stop(self):
        """Disables the cpuload monitor"""
        if self._monitor is None: return
        self._gdb.execute("set cpuload_monitor_all_count = 0")

    @cached_property
    def sample(self) -> tuple[int, int, dict[int, tuple[int, int]]]:
        """
        Samples the cpuload monitor and computes the difference to the last
        sample.
        :return: a tuple of (start time since enabled [µs], interval from last
                 sample [µs], tasks dict[tcb [ptr] -> (total runtime [µs], difference
                 from last sample [µs])]).
        """
        if self._system_load is None or not self._system_load["initialized"]:
            return (0, 0, {})

        sl = self._load()
        # compute the delta running times
        sample = {}
        for tcb, total in sl["tasks"].items():
            ptotal = self._psl["tasks"].get(tcb, total)
            sample[tcb] = (total, total - ptotal)
        interval = self._device.uptime - self._psl["hrt"]
        # remember the new system load as previous load
        self._psl = sl
        return (sl["start"], interval, sample)


_SYSTEM_LOAD = None
def system_load(gdb):
    """
    :return: The SystemLoad singleton object.
    """
    global _SYSTEM_LOAD
    if _SYSTEM_LOAD is None:
        _SYSTEM_LOAD = SystemLoad(gdb)
    return _SYSTEM_LOAD


def restart_system_load_monitor(gdb):
    """
    Starts the system load monitor if not started, else resets the sample
    interval to right now.
    """
    system_load(gdb).sample
    system_load(gdb).restart()



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
import os
import shutil
import subprocess
import hashlib
from pathlib import Path
from contextlib import contextmanager
import logging
LOGGER = logging.getLogger("patch:ops")

def _repopath(path: str = None) -> Path:
    return Path(__file__).parents[1] / (path or "")

class OperationError(Exception):
    """Base exception for patching-related errors."""
    pass

# -----------------------------------------------------------------------------
class CopyOperation:
    """
    Copies a file from the store to the repository and removes it again.

    In case the destination location already exists, the file is stored in a
    cache, then overwritten. On restore, the cached file is copied back
    Note that you can set `src` to `None` if you want to remove a file inside
    the repository.
    """
    PATCH_STORE = _repopath(".patch_store")

    def __init__(self, source: Path, destination: Path):
        """
        :param source: File to copy into the repository. If set to `None`, the
                       destination file is simply removed.
        :param destination: File to add or overwrite. In case, the file exists,
                            it is copied to the `.patch_store` folder from where
                            it is copied back when restoring.
        """
        self._src = None
        self._dst = Path(destination)
        self._hash = str(self._dst.absolute())
        if source is not None:
            self._src = Path(source)
            self._hash += str(self._src.absolute())
        self._hash = hashlib.md5(self._hash.encode("utf-8")).hexdigest()
        self._store = self.PATCH_STORE / str(self._hash)
        self._dstrestore = self._store / self._dst.name

    def do(self) -> bool:
        """
        Copies the `source` file to the `destination` location.
        If the destination exists, it is copied to the `.patch_store`, before
        getting overwritten by the source.

        :return: `True` if applied successfully.
        """
        if self._dst.exists():
            LOGGER.debug(f"Caching {self._dst} -> {self._dstrestore}")
            self._store.mkdir(parents=True, exist_ok=True)
            shutil.copy2(self._dst, self._dstrestore)
        if self._src is not None:
            LOGGER.debug(f"Copying {self._src} -> {self._dst}")
            self._dst.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(self._src, self._dst)
        return True

    def undo(self) -> bool:
        """
        Removes the source file and copies the original destination file if it
        is found in the `.patch_store` folder.

        :return: `True` if restored successfully.
        """
        LOGGER.debug(f"Removing {self._dst}")
        self._dst.unlink(missing_ok=True)
        if self._dstrestore.exists():
            LOGGER.debug(f"Uncaching {self._dstrestore} -> {self._dst}")
            shutil.copy2(self._dstrestore, self._dst)
            self._dstrestore.unlink()
        return True

    def __repr__(self) -> str:
        if self._src is None:
            return f"Copy(Store <- {self._dst.name})"
        if self._src.name == self._dst.name:
            return f"Copy({self._dst.name})"
        return f"Copy({self._src.name} -> {self._dst.name})"


class PatchOperation:
    """
    Applies a patch in unified diff format to a directory.
    You can generate the patch from a git commit: `git format-patch -1 <sha>`.
    However, there are many more methods that generate unified diffs.
    """
    def __init__(self, root: Path, patch: Path):
        """
        :param root: directory to relocate the relative paths inside the patch to.
        :param patch: path to patch in unified diff format.

        :raises ValueError: if patch does not exists, or it cannot be parsed.
        """
        self._root = Path(root).absolute().resolve()
        self._patch = Path(patch).absolute().resolve()
        self._cmd = "patch --strip=1 --forward --ignore-whitespace --reject-file=- " \
                    f"--posix --directory={self._root} --input={self._patch} --silent"
        self._cmd_rev = self._cmd + " --reverse"
        self._cmd_check = " --dry-run"

    def do(self) -> bool:
        """
        Applies the patch to the directory.

        :return: `True` if applied successfully.
        """
        LOGGER.debug(f"Applying patch {self._patch}")
        cmd_check = self._cmd + self._cmd_check
        if subprocess.run(cmd_check, shell=True).returncode:
            LOGGER.debug(cmd_check)
            return False
        LOGGER.debug(self._cmd)
        return not subprocess.run(self._cmd, shell=True).returncode

    def undo(self) -> bool:
        """
        Applies the reversed patch to the directory.

        :return: `True` if restored successfully.
        """
        LOGGER.debug(f"Reverting patch {self._patch}")
        cmd_check = self._cmd_rev + self._cmd_check
        if subprocess.run(cmd_check, shell=True).returncode:
            LOGGER.debug(cmd_check)
            return False
        LOGGER.debug(self._cmd_rev)
        return not subprocess.run(self._cmd_rev, shell=True).returncode

    def __repr__(self) -> str:
        return f"Patch({self._patch.name})"


# -----------------------------------------------------------------------------
class PatchManager:
    """
    Stores a number of named operations and applies them all.
    """
    def __init__(self, name: str, ops: list[CopyOperation|PatchOperation]):
        """
        :param name: Short, human-readable name of patch.
        :param ops: list of operations.
        """
        self.name = name
        # make sure that PatchOperations happen first, since they can fail
        self._ops = sorted(ops, key=lambda o: isinstance(o, CopyOperation))

    def do(self):
        """
        Runs all operations to apply the patch. All operations are attempted
        even if any of them fail.

        :raises `OperationError`: if any of the operations failed.
        """
        LOGGER.info(f"Applying '{self.name}'")
        for op in self._ops:
            if not op.do():
                raise OperationError(f"Patching failed: {op}")

    def undo(self):
        """
        Runs all operations to apply the patch. All operations are attempted
        even if any of them fail.

        :raises `OperationError`: if any of the operations failed.
        """
        LOGGER.info(f"Restoring '{self.name}'")
        for op in self._ops:
            if not op.undo():
                raise OperationError(f"Reverting failed: {op}")

    @contextmanager
    def apply(self):
        """
        Context manager version to apply and restore the patch within a scope.

        ```py
        # patch repository
        with patch.apply():
            # run tests here
        # restore repository
        ```
        """
        try:
            self.do()
            yield
        finally:
            self.undo()



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
# Test Benches

A test bench uses a specific combination of pre-configured modules.
Connecting to a test bench uses the Python GDB with all scripts imported.


"""

from .fmu import debug as fmu, shell
from .skynode import debug as skynode



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
from functools import cached_property
from . import utils
from .system_load import system_load
from .device import Device
from .base import Base
from dataclasses import dataclass
from collections import defaultdict
import rich.box
from rich.text import Text
from rich.table import Table

# The mapping from register name to position on the thread stack
_XCP_REGS_MAP: dict[str, int] = {
    "msp":     0,
    "basepri": 1,
    "r4":      2,
    "r5":      3,
    "r6":      4,
    "r7":      5,
    "r8":      6,
    "r9":      7,
    "r10":     8,
    "r11":     9,
    "s16":     11 + 0,
    "s17":     11 + 1,
    "s18":     11 + 2,
    "s19":     11 + 3,
    "s20":     11 + 4,
    "s21":     11 + 5,
    "s22":     11 + 6,
    "s23":     11 + 7,
    "s24":     11 + 8,
    "s25":     11 + 9,
    "s26":     11 + 10,
    "s27":     11 + 11,
    "s28":     11 + 12,
    "s29":     11 + 13,
    "s30":     11 + 14,
    "s31":     11 + 15,
    "r0":      27 + 0,
    "r1":      27 + 1,
    "r2":      27 + 2,
    "r3":      27 + 3,
    "r12":     27 + 4,
    "r14":     27 + 5,
    "r15":     27 + 6,
    "xpsr":    27 + 7,
    "s0":      27 + 8,
    "s1":      27 + 9,
    "s2":      27 + 10,
    "s3":      27 + 11,
    "s4":      27 + 12,
    "s5":      27 + 13,
    "s6":      27 + 14,
    "s7":      27 + 15,
    "s8":      27 + 16,
    "s9":      27 + 17,
    "s10":     27 + 18,
    "s11":     27 + 19,
    "s12":     27 + 20,
    "s13":     27 + 21,
    "s14":     27 + 22,
    "s15":     27 + 23,
    "fpscr":   27 + 24,
}

class Task(Base):
    """
    NuttX task
    """
    _STACK_COLOR = 0xdeadbeef
    _FILE_DESCRIPTORS_PER_BLOCK = 6 # TODO query from ELF

    @dataclass
    class Load:
        total: int
        """Total task runtime in µs"""
        interval: int
        """Interval task runtime in µs since last sample"""
        delta: int
        """Task runtime in µs within the interval"""

        @property
        def relative(self) -> float:
            """Relative runtime within the interval"""
            return self.delta / self.interval if self.interval else 0

    def __init__(self, gdb, tcb_ptr: "gdb.Value"):
        super().__init__(gdb)
        self._tcb = tcb_ptr
        self._system_load = system_load(self._gdb)
        self.pid = self._tcb["pid"]
        self.init_priority = int(self._tcb["init_priority"])
        self.stack_limit = int(self._tcb["adj_stack_size"])
        self.stack_ptr = self._tcb["stack_base_ptr"]
        self._is_running_switched = None

    @cached_property
    def name(self) -> str:
        """Name of the task"""
        try:
            return self._tcb["name"].string()
        except:
            return "?"

    @cached_property
    def sched_priority(self) -> int:
        """The scheduled priority of the task"""
        return int(self._tcb["sched_priority"])

    @cached_property
    def _statenames(self):
        return self._gdb.types.make_enum_dict(self._gdb.lookup_type("enum tstate_e"))

    @cached_property
    def state(self) -> str:
        """Task state name"""
        for name, value in self._statenames.items():
            if value == self._tcb["task_state"]:
                return name
        return "UNKNOWN"

    @cached_property
    def short_state(self) -> str:
        """The short name of the task state"""
        mapping = {
            "TASK_PENDING": "PEND", "TASK_READYTORUN": "READY", "TASK_RUNNING": "RUN",
            "WAIT_SEM": "w:sem", "WAIT_SIG": "w:sig"}
        return mapping.get(self.state.replace("TSTATE_", ""), "???")

    def _is_state_in(self, *states: list[str]) -> bool:
        states = {int(self._statenames[s]) for s in states}
        if int(self._tcb["task_state"]) in states:
            return True
        return False

    @cached_property
    def is_waiting(self) -> bool:
        """The task is waiting for a semaphore or signal"""
        if self._is_state_in("TSTATE_WAIT_SEM", "TSTATE_WAIT_SIG"):
            return True

    @cached_property
    def is_runnable(self) -> bool:
        """The task is pending, ready to run, or running"""
        return self._is_state_in("TSTATE_TASK_PENDING",
                                 "TSTATE_TASK_READYTORUN",
                                 "TSTATE_TASK_RUNNING")

    @cached_property
    def stack_used(self) -> int:
        """The amount of stack used by the thread in bytes"""
        if not self.stack_ptr: return 0
        stack_u32p = self.stack_ptr.cast(self._gdb.lookup_type("unsigned int").pointer())
        # Stack grows from top to bottom, we do a binary search for the
        # 0xdeadbeef value from the bottom upwards
        watermark = utils.binary_search_last(stack_u32p, self._STACK_COLOR, hi=self.stack_limit // 4) + 1
        # validate the binary search (does not seem necessary)
        # for ii in range(0, watermark):
        #     if stack_u32p[ii] != self._STACK_COLOR:
        #         print(f"{self.name}: Correcting stack size from {watermark * 4} to {ii * 4}!")
        #         return ii * 4
        return self.stack_limit - watermark * 4

    @cached_property
    def waiting_for(self) -> str:
        """
        What the task is waiting for. If its a semaphore, return an object,
        otherwise a string.
        """
        if self._is_state_in("TSTATE_WAIT_SEM"):
            from .semaphore import Semaphore
            sem = self._tcb['waitsem']
            ostr = f"{int(sem):#08x} "
            if descr := self.description_at(sem): ostr += f"<{descr}> "
            ostr += Semaphore(self._gdb, sem).to_string()
            return ostr
        if self._is_state_in("TSTATE_WAIT_SIG"):
            return "signal"
        return ""

    @cached_property
    def files(self) -> list["gdb.Value"]:
        """The list of inode pointers the task holds"""
        filelist = self._tcb["group"]["tg_filelist"]
        rows = filelist["fl_rows"]
        files = filelist["fl_files"]
        result = []
        for ri in range(rows):
            for ci in range(self._FILE_DESCRIPTORS_PER_BLOCK):
                file = files[ri][ci]
                if inode := file["f_inode"]:
                    result.append(file)
        return result

    @cached_property
    def location(self) -> "gdb.Block":
        """The block the task is currently executing"""
        if self.is_current_task:
            pc = self.read_register("pc")
        else:
            pc = self._tcb["xcp"]["regs"][32]
        block = self.block(pc)
        while block and not block.function:
            block = block.superblock
        return block.function if block else int(pc)

    def switch_to(self) -> bool:
        """Switch to this task by writing the register file"""
        if self.is_current_task:
            return False
        regs = {name: self._tcb["xcp"]["regs"][offset]
                for name, offset in _XCP_REGS_MAP.items()}
        self.write_registers(regs)
        self._is_running_switched = True
        return True

    def switch_from(self) -> dict[str, int] | None:
        """Switch to this task by writing the register file"""
        if not self.is_current_task:
            return None
        self._is_running_switched = False
        if self.short_state == "RUN":
            self._is_running_switched = None
        return self.registers

    @cached_property
    def is_current_task(self) -> bool:
        """If the task is currently running"""
        if self._is_running_switched is not None:
            return self._is_running_switched
        return self.short_state == "RUN"

    @cached_property
    def load(self) -> Load:
        """The task load based on the system load monitor"""
        if self._system_load is None: return self.Load(0, 0, 0)
        _, interval, sl = self._system_load.sample
        total, delta = sl.get(int(self._tcb), (0,0))
        return self.Load(total, interval, delta)

    def __repr__(self) -> str:
        return f"Task({self.name}, {self.pid})"

    def __str__(self) -> str:
        ostr = self.__repr__() + f": state={self.state}, "
        ostr += f"prio={self.sched_priority}({self.init_priority}), "
        ostr += f"stack={self.stack_used}/{self.stack_limit}"
        if waiting := self.waiting_for:
            ostr += f", waiting={waiting}"
        return ostr


def all_tasks(gdb) -> list[Task]:
    """Return a list of all tasks"""
    type_tcb_s = gdb.lookup_type("struct tcb_s").pointer()
    def _tasks(name):
        tcbs = []
        if (task_list := gdb.lookup_global_symbol(name)) is None:
            return []
        task_list = task_list.value()
        current_task = task_list["head"]
        if current_task:
            while True:
                tcbs.append(Task(gdb, current_task.cast(type_tcb_s)))
                if current_task == task_list["tail"]:
                    break
                next_task = current_task["flink"]
                # if next_task["blink"] == current_task:
                #   LOGGER.error(f"Task linkage is broken in {tasks}!")
                #   break
                current_task = next_task
        return tcbs

    tcbs  = _tasks("g_pendingtasks")
    tcbs += _tasks("g_readytorun")
    tcbs += _tasks("g_waitingforsemaphore")
    tcbs += _tasks("g_waitingforsignal")
    tcbs += _tasks("g_inactivetasks")
    return tcbs


_RESTORE_REGISTERS = None
def task_switch(gdb, pid: int) -> bool:
    """
    Switch to another task.
    On initial switch the current register file is saved and can be restored
    by passing a PID <0.

    :param pid: the PID of the task to switch to, or <0 to restore initial task.
    :return: Success of switching operation.
    """
    global _RESTORE_REGISTERS
    # Restore registers to original task
    if pid < 0:
        if _RESTORE_REGISTERS is not None:
            Base(gdb).write_registers(_RESTORE_REGISTERS)
        _RESTORE_REGISTERS = None
        return True

    # Otherwise find the new pointer
    tcbs = all_tasks(gdb)
    if (next_task := next((t for t in tcbs if int(t.pid) == pid), None)) is not None:
        # Find the currently executing task and save their registers
        if (current_task := next((t for t in tcbs if t.is_current_task), None)) is not None:
            regs = current_task.switch_from()
        else:
            regs = Base(gdb).registers
        # We only care about the first register set for restoration
        if next_task.switch_to():
            if _RESTORE_REGISTERS is None:
                _RESTORE_REGISTERS = regs
            print(f"Switched to task '{next_task.name}' ({pid}).")
            return True

        print("Task already loaded!")
        return False

    print(f"Unknown task PID '{pid}'!")
    return False


def all_tasks_as_table(gdb, sort_key: str = None, with_stack_usage: bool = True,
                       with_file_names: bool = True, with_waiting: bool = True) \
                                    -> tuple[Table, str] | tuple[None, None]:
    """
    Return a table of running tasks similar to the NSH top command.

    :param sort_key: optional lambda function to sort the table rows.
    :param with_stack_usage: compute and show the task stack usage.
    :param with_file_names: show what files the task has open.
    :param with_waiting: show what the task is waiting for.

    :return: The task table and additional output. If no tasks are found, return `None`.
    """
    table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
    table.add_column("struct tcb_s*", justify="right", no_wrap=True)
    table.add_column("pid", justify="right", no_wrap=True)
    table.add_column("Task Name")
    table.add_column("Location", no_wrap=True)
    table.add_column("CPU(ms)", justify="right")
    table.add_column("CPU(%)", justify="right")
    table.add_column("Stack\nUsage", justify="right")
    table.add_column("Avail\nStack", justify="right")
    table.add_column("Prio", justify="right")
    table.add_column("Base", justify="right")
    if with_file_names: table.add_column("Open File Names")
    else: table.add_column("FDs", justify="right")
    table.add_column("State")
    if with_waiting: table.add_column("Waiting For")

    tasks = all_tasks(gdb)
    if not tasks: return None, None
    interval_us = tasks[0].load.interval
    if not interval_us:
        start, *_ = tasks[0]._system_load.sample
        total_interval_us = (Device(gdb).uptime - start) or 1
    total_user_us, total_idle_us = 0, 0
    rows = []
    for task in tasks:
        # Remember the CPU loads for idle task and add it for the other tasks
        if task.pid == 0: total_idle_us = task.load.delta if interval_us else task.load.total
        else: total_user_us += task.load.delta if interval_us else task.load.total

        # Find the file names or just the number of file descriptors
        if with_file_names:
            file_names = [task.read_string(f["f_inode"]["i_name"]) for f in task.files]
            file_description = ", ".join(sorted(list(set(file_names))))
        else:
            file_description = len(task.files)

        # Add all the values per row
        relative = task.load.relative if interval_us else task.load.total / total_interval_us
        stack_overflow = with_stack_usage and task.stack_used >= task.stack_limit
        row = [hex(task._tcb), task.pid, task.name,
               hex(task.location) if isinstance(task.location, int) else task.location.name,
               task.load.total//1000, f"{(relative * 100):.1f}",
               Text.assemble((str(task.stack_used) if with_stack_usage else "", "bold red" if stack_overflow else "")),
               Text.assemble((str(task.stack_limit), "bold" if stack_overflow else "")),
               Text.assemble((str(task.sched_priority), "bold red" if task.sched_priority > task.init_priority else "")),
               task.init_priority, file_description, task.short_state]
        if with_waiting:
            row.append(task.waiting_for)
        rows.append(row)

    # Sort the rows by PID by default and format the table
    for row in sorted(rows, key=lambda l: l[1] if sort_key is None else sort_key):
        table.add_row(*[r if isinstance(r, Text) else str(r) for r in row],
                      style="bold blue" if row[11] == "RUN" else None)

    # Add the task information
    running = sum(1 for t in tasks if t.is_runnable)
    sleeping = sum(1 for t in tasks if t.is_waiting)
    output = f"Processes: {len(tasks)} total, {running} running, {sleeping} sleeping\n"

    # Add CPU utilization and guard against division by zero
    if not interval_us: interval_us = total_interval_us
    user = 100 * total_user_us / interval_us;
    idle = 100 * total_idle_us / interval_us;
    sched = 100 * (interval_us - total_idle_us - total_user_us) / interval_us;
    output += f"CPU usage: {user:.1f}% tasks, {sched:.1f}% sched, {idle:.1f}% idle\n"

    # Uptime finally
    output += f"Uptime: {Device(gdb).uptime/1e6:.2f}s total, {interval_us/1e6:.2f}s interval\n"
    return table, output


def all_files_as_table(gdb, sort_key: str = None) -> Table | None:
    """
    Return a table of open files owned by tasks.

    :param sort_key: optional lambda function to sort the table rows.

    :return: The file table or `None` if no tasks exist.
    """
    tasks = all_tasks(gdb)
    if not tasks: return None

    files = {}
    file_tasks = defaultdict(set)
    for task in tasks:
        # Find the file names or just the number of file descriptors
        for f in task.files:
            files[int(f["f_inode"])] = f["f_inode"]
            file_tasks[int(f["f_inode"])].add(task)
    # Format the rows
    rows = []
    for addr, inode in files.items():
        rows.append((hex(addr),
                     hex(inode["i_private"]) if inode["i_private"] else "",
                     task.read_string(inode["i_name"]),
                     ", ".join(sorted(t.name for t in file_tasks[addr]))))
    # sort and add rows
    table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
    table.add_column("struct inode*", justify="right", no_wrap=True)
    table.add_column("i_private*", justify="right", no_wrap=True)
    table.add_column("Name")
    table.add_column("Tasks")
    for row in sorted(rows, key=lambda l: l[2] if sort_key is None else sort_key):
        table.add_row(*row)
    return table



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
.. include:: callgraph.md
"""

from __future__ import annotations
import re, os
import itertools
import logging
from pathlib import Path
from collections import defaultdict
from .backtrace import Backtrace
from .utils import read_gdb_log
LOGGER = logging.getLogger(__name__)


def callgraph_from_backtrace(backtraces: list[str],
                             BacktraceClass: Backtrace = None,
                             output_graphviz: Path = None,
                             output_pyvis: Path = None):
    """
    Convert a GDB backtrace log file into a dot, svg and pyvis file.

    :param backtraces: list of strings each containing a backtrace.
    :param BacktraceClass: One of the classes of `emdbg.analyze.backtrace`.
    :param output_graphviz: Output path of the graphviz file (`.dot` suffix).
        If you set its suffix to `.svg`, the `dot` command is used to generate
        a SVG file instead.
    :param output_pyvis: Output path to a pyvis file. (Requires the `pyvis`
        module to be installed).
    """

    if BacktraceClass is None:
        BacktraceClass = Backtrace

    backts = defaultdict(set)
    for description in backtraces:
        bt = BacktraceClass(description)
        if bt.is_valid:
            backts[bt.type].add(bt)
            # if bt.type == "unknown":
            #     print(bt)
            #     print(bt.description)
        else:
            LOGGER.error(bt)
            LOGGER.error(bt.description)

    nodes = {}
    edges = defaultdict(int)

    def itertools_pairwise(iterable):
        a, b = itertools.tee(iterable)
        next(b, None)
        return zip(a, b)

    for btype, bts in backts.items():
        for bt in bts:
            for frame in bt.frames:
                nodes[frame._unique_location] = frame
            for f1, f2 in itertools_pairwise(bt.frames):
                edges[(f2._unique_location, f1._unique_location, str((btype or "").lower()))] += 1

    sources = set(nodes)
    sinks = set(nodes)
    max_calls = max(edges.values())
    for source, sink, _ in edges.keys():
        sources.discard(sink)
        sinks.discard(source)

    def _n(name):
        return re.sub(r"[, :<>]", "_", name)

    if output_pyvis:
        import pyvis
        net = pyvis.network.Network(height="100%", width="100%", select_menu=True)
        for node in sorted(nodes):
            kwargs = {"label": node}
            if node in sinks:
                kwargs.update({"borderWidth": 3, "color": "LightBlue"})
            elif node in sources:
                kwargs.update({"borderWidth": 3, "color": "LightGreen"})
            net.add_node(_n(node), label=node)
        for edge in sorted(edges):
            net.add_edge(_n(edge[0]), _n(edge[1]), label=edge[2], arrows="to")
        net.toggle_physics(True)
        net.show(output_pyvis, notebook=False)

    if output_graphviz:
        output_graphviz = Path(output_graphviz)
        import graphviz
        dot = graphviz.Digraph()
        for node in sorted(nodes):
            frame = nodes[node]
            kwargs = {
                "label": f"{frame.function}:{frame.line}",
                "URL": f"subl://open?url={Path(frame.filename).absolute()}&line={frame.line}",
            }
            for pattern, style in BacktraceClass.COLORS.items():
                if re.search(pattern, node):
                    kwargs.update(style)
            if node in sinks:
                kwargs.update({"style": "bold,filled", "fillcolor": "LightBlue"})
            elif node in sources:
                kwargs.update({"style": "bold,filled", "fillcolor": "LightGreen"})
            dot.node(_n(node), **kwargs)
        for edge in sorted(edges):
            dot.edge(_n(edge[0]), _n(edge[1]), label=edge[2] or str(edges[edge]),
                     penwidth=str(max(edges[edge]/max_calls * 10, 0.5)))
        output_dot = output_graphviz.with_suffix(".dot")
        output_dot.write_text(dot.source)
        if output_graphviz.suffix == ".svg":
            os.system(f"dot -Tsvg -o {output_graphviz} {output_dot}")
            os.system(f"rm {output_dot}")


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    import argparse
    from .backtrace import *

    parser = argparse.ArgumentParser(description="Backtrace Analyzer")
    parser.add_argument(
        "file",
        type=Path,
        help="The GDB log containing the backtraces.")
    parser.add_argument(
        "--graphviz",
        type=Path,
        help="The file to render the dot graph in.")
    parser.add_argument(
        "--svg",
        action="store_true",
        default=False,
        help="Render into SVG file using the same name as the input.")
    parser.add_argument(
        "--pyvis",
        type=Path,
        help="The file to render the pyvis graph in.")
    values = {
        "FileSystem": FileSystemBacktrace,
        "SPI": SpiBacktrace,
        "I2C": I2cBacktrace,
        "CAN": CanBacktrace,
        "UART": UartBacktrace,
        "Semaphore": SemaphoreBacktrace,
        "Generic": Backtrace,
    }
    parser.add_argument(
        "--type",
        choices=values.keys(),
        help="The backtrace class to use.")
    args = parser.parse_args()
    BacktraceClass = values.get(args.type)

    if BacktraceClass is None:
        if "_sdmmc" in args.file.name:
            BacktraceClass = FileSystemBacktrace
        elif "_spi" in args.file.name:
            BacktraceClass = SpiBacktrace
        elif "_i2c" in args.file.name:
            BacktraceClass = I2cBacktrace
        elif "_can" in args.file.name:
            BacktraceClass = CanBacktrace
        elif "_uart" in args.file.name:
            BacktraceClass = UartBacktrace
        elif "_semaphore" in args.file.name:
            BacktraceClass = SemaphoreBacktrace
        else:
            BacktraceClass = Backtrace

    graphviz = args.graphviz
    if args.svg:
        graphviz = Path(str(args.file.with_suffix(".svg")).replace("calltrace_", "callgraph_"))

    backtraces = re.split(r"(?:Breakpoint|Hardware .*?watchpoint) \d", read_gdb_log(args.file)[20:])
    callgraph_from_backtrace(backtraces, BacktraceClass,
                             output_graphviz=graphviz, output_pyvis=args.pyvis)



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
from serial.tools import list_ports
import logging
LOGGER = logging.getLogger("serial")


class SerialException(Exception):
    """Exception for all serial port related errors."""
    pass


def find_serial_port(identifier=None):
    """
    Finds the serial port with the `identifier` serial number.
    If not found, an exception is raised with all discovered serial ports in
    the message, or a list of serial ports, if `identifier` is None.

    :raises `SerialException`: if no serial port matches `identifier`.
    :return:
    - The serial port whose serial number matches the `identifier`, or
    - A list of all serial ports if `identifier` is `None`.
    """
    serials = []
    for port in list_ports.comports():
        if not identifier:
            LOGGER.debug(f"Using first found serial port '{port.serial_number}'.")
            return port
        if port.serial_number == identifier:
            return port
        serials.append(port)

    if identifier is None:
        serials = "\n\t- ".join(s.serial_number for s in serials)
        msg = f"Unable to find '{identifier or 'any'}' serial port!\n"
        msg += f"Available identifiers are:\n\t- {serials}\n"
        raise SerialException(msg)

    return serials



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
from .base import Base


class UartBuffer(Base):
    """
    Pretty Printing UART buffers
    """

    def __init__(self, gdb, buf_ptr: "gdb.Value"):
        super().__init__(gdb)
        self._buf = buf_ptr

    def children(self) -> Iterator[tuple[str, Any]]:
        size, head, tail = self._buf['size'], self._buf['head'], self._buf['tail']
        used = head - tail if (tail <= head) else size - tail + head
        free = size - used
        yield ("sem", self._buf['sem'])
        yield ("size", size)
        yield ("buffer", self._buf['buffer'])
        yield ("tail -> head", f"{tail} -> {head}: {used} used, {free} free")
        if used:
            if (tail < head):
                # [tail, head]
                content = self.read_memory(self._buf['buffer'] + tail, used)
            else:
                # head], [tail
                content = self.read_memory(self._buf['buffer'] + tail, size - tail)
                content += self.read_memory(self._buf['buffer'], head)
            # Convert to Prints else \hh hex values
            content = "".join(chr(v) if chr(v).isprintable() else f"\\{v:02x}"
                              for v in content.tobytes())
            yield ("content", content)



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
import re
import time
import logging
from contextlib import contextmanager
from pathlib import Path
from serial import Serial
from serial.threaded import ReaderThread, Protocol
from .utils import find_serial_port
from ..utils import add_datetime as add_dt

LOGGER = logging.getLogger("serial:nsh")


class _NshReader(Protocol):
    def __init__(self, device: Serial):
        super().__init__()
        self.stream = ""
        self.device = device
        self._data_received = lambda _: None

    def data_received(self, data: bytes):
        new_stream = data.decode("utf-8", errors="replace")
        self._data_received(new_stream)
        self.stream += new_stream

    def read_packets(self, separator: str) -> list[str]:
        if separator not in self.stream:
            return []
        *packets, self.stream = self.stream.split(separator)
        return packets

    def write_line(self, line: str):
        # Pace the transmission to 1 char / 10ms to not loose character
        for char in (line + "\n"):
            self.device.write(char.encode("utf-8"))
            # self.device.flushOutput()
            time.sleep(0.01)

    def clear_input(self):
        self.device.reset_input_buffer()
        self.stream = ""

    def clear_output(self):
        self.device.reset_output_buffer()

    def clear(self):
        self.clear_output()
        self.clear_input()


# -----------------------------------------------------------------------------
class Nsh:
    """
    Manages the NSH protocol, in particular, receiving data in the background
    and logging it out to the INFO logger.
    Several convenience methods allow you to send a command and receive its
    response, or wait for a certain pattern to arrive in the stream.

    Transmitting a command is intentionally slow to not overflow the NSH receive
    buffers while it is being debugged. Note that the PX4 target needs to be
    running for the NSH to be functional.
    """
    _ANSI_ESCAPE = re.compile(r"\x1b(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
    _NEWLINE = "\r\n"
    _PROMPT = "nsh> "
    _TIMEOUT = 3

    def __init__(self, reader: Protocol):
        """
        Use the `nsh` context manager to build this class correctly.

        :param reader: The background thread reader protocol.
        """
        self._serial = reader
        self._serial._data_received = self._print
        self._print_data = ""
        self.filter_ansi_escapes = True
        self._logfile = None

    def _write_line(self, line):
        self._serial.write_line(line)

    def _print(self, data: str):
        self._print_data += data
        if Nsh._NEWLINE in self._print_data:
            *lines, self._print_data = self._print_data.split(Nsh._NEWLINE)
            for line in self._filter(lines):
                LOGGER.info(line)
                if self._logfile is not None:
                    self._logfile.write(line + "\n")

    def _filter(self, lines):
        if self.filter_ansi_escapes:
            lines = [self._ANSI_ESCAPE.sub('', line) for line in lines]
        return lines

    def _read_packets(self, separator: str, timeout: float = _TIMEOUT) -> list[str]:
        start = time.time()
        while True:
            if packets := self._serial.read_packets(separator):
                return packets
            if time.time() - start > timeout:
                break
            time.sleep(0.1)
        return []

    def clear(self):
        """Clear the receive and transmit buffers."""
        self._serial.clear()
        self._print_data = ""

    def log_to_file(self, filename: Path | str, add_datetime: bool = False):
        """
        Log the received data to `filename` or close the log file.

        :param filename: file to log into. If `None`, the file is closed and
                         logging is disabled.
        :param add_datetime: appends the date and time to the filename.
                             See `emdbg.utils.add_datetime`

        :return: the actual filename with date and time (if selected).
        """
        if self._logfile is not None:
            self._logfile.close()

        if filename is None:
            self._logfile = None
            return None

        filename = add_dt(filename) if add_datetime else Path(filename)
        self._logfile = filename.open("wt")
        return filename


    def read_lines(self, timeout: float = _TIMEOUT) -> list[str]:
        """
        Return any lines received within `timeout`.
        Note that any ANSI escape codes (for color or cursor position) are
        filtered out.

        :param timeout: seconds to wait until new lines arrive.

        :return: list of received lines or an empty list
        """
        return self._filter(self._read_packets(Nsh._NEWLINE, timeout))


    def wait_for(self, pattern: str, timeout: float = _TIMEOUT) -> list[str]:
        """
        Waits for a regex pattern to appear in a line in the stream.
        This function reads any new received lines and searches for `pattern` in
        every line. If the line matches, all lines are returned.
        Note that any ANSI escape codes (for color or cursor position) are
        filtered out.

        :param pattern: regex pattern to search for via `re.search`. To match
                        line beginnings and ends you must use `^pattern$`.
        :param timeout: seconds to wait until new lines arrive.

        :return: list of received lines until matched pattern or an empty list
        """
        lines = []
        start = time.time()
        while True:
            if time.time() - start > timeout:
                break
            new_lines = self.read_lines(0)
            lines.append(new_lines)
            for line in new_lines:
                if re.search(pattern, line):
                    return lines
            time.sleep(0.1)
        LOGGER.warning(f"Waiting for '{pattern}' timed out after {timeout:.1f}s!")
        return []

    def wait_for_prompt(self, timeout: float = _TIMEOUT) -> list[str]:
        """
        Waits to the `nsh> ` prompt to arrive in the stream.
        Note that any ANSI escape codes (for color or cursor position) are
        filtered out.

        :param timeout: seconds to wait until the prompt arrives.

        :return: all lines until the prompt arrives.
        """
        if prompts := self._read_packets(Nsh._NEWLINE + Nsh._PROMPT, timeout):
            prompt = Nsh._PROMPT + Nsh._PROMPT.join(prompts)
            return self._filter(prompt.split(Nsh._NEWLINE))
        LOGGER.warning(f"Waiting for 'nsh> ' prompt timed out after {timeout:.1f}s!")
        return []


    def command(self, command: str, timeout: float = _TIMEOUT) -> list[str]:
        """
        Send a command and return all lines until the next prompt.
        If the command is asynchronous, you need to poll for new lines separately.
        Note that any ANSI escape codes (for color or cursor position) are
        filtered out.

        :param command: command string to send to the NSH.
        :param timeout: seconds to wait until the prompt arrives.

        :return: all lines from the command issue until the next prompt arrives.
        """
        self._serial.clear()
        self._write_line(command)
        if timeout is not None:
            return self.wait_for_prompt(timeout)

    def command_nowait(self, command: str):
        """
        Send a command to the NSH without waiting for a response.

        :param command: command string to send to the NSH.
        """
        self.command(command, None)


    def reboot(self, timeout: int = 15) -> list[str]:
        """
        Send the reboot command and wait for the reboot to be completed.

        :param timeout: seconds to wait until the prompt arrives.

        :return: all lines from reboot until the next prompt.
        """
        return self.command("reboot", timeout)

    def is_alive(self, timeout: float = _TIMEOUT, attempts: int = 4) -> bool:
        """
        Check if the NSH is responding to newline inputs with a `nsh> ` prompt.
        The total timeout is `attempts * timeout`!

        :param timeout: seconds to wait until the prompt arrives.
        :param attempts: number of times to send a newline and wait.
        :return: `True` is NSH responds, `False` otherwise
        """
        self._serial.clear()
        attempt = 0
        timeout = timeout / attempts
        while attempt < attempts:
            self._write_line("")
            if self.wait_for_prompt(timeout):
                return True
            attempt += 1
        return False


# -----------------------------------------------------------------------------
@contextmanager
def nsh(serial: str, baudrate: int = 57600):
    """
    Opens a serial port with the `serial` number and closes it again.

    :param serial: the serial number of the port to connect to.
    :param baudrate: the baudrate to use.

    :raises `SerialException`: if serial port is not found.

    :return: yields an initialized `Nsh` object.
    """
    nsh = None
    port = find_serial_port(serial)
    try:
        LOGGER.info(f"Starting on port '{serial}'..." if serial else "Starting...")
        device = Serial(port.device, baudrate=baudrate, timeout=0.01)
        with ReaderThread(device, lambda: _NshReader(device)) as reader:
            nsh = Nsh(reader)
            yield nsh
    finally:
        if nsh is not None: nsh.log_to_file(None)
        LOGGER.debug("Stopping.")



# Copyright (c) 2015, Gallopsled et al.
# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: MIT
# Adapted from https://github.com/Gallopsled/pwntools

from __future__ import annotations
from threading import Event
from contextlib import contextmanager
import time
from .gdb import Interface
import logging
LOGGER = logging.getLogger(__name__)
from ...logger import VERBOSITY

__all__ = ["Gdb"]

class Breakpoint:
    """Mirror of `gdb.Breakpoint` class.

    See https://sourceware.org/gdb/onlinedocs/gdb/Breakpoints-In-Python.html
    for more information.
    """
    def __init__(self, conn, *args, **kwargs):
        """Do not create instances of this class directly.

        Use `Gdb.Breakpoint` instead.
        """
        # Creates a real breakpoint and connects it with this mirror
        self.conn = conn
        self.server_breakpoint = conn.root.set_breakpoint(
            self, hasattr(self, 'stop'), *args, **kwargs)

    def __getattr__(self, item):
        """Return attributes of the real breakpoint."""
        if item in (
                '____id_pack__',
                '__name__',
                '____conn__',
                'stop',
        ):
            # Ignore RPyC netref attributes.
            # Also, if stop() is not defined, hasattr() call in our
            # __init__() will bring us here. Don't contact the
            # server in this case either.
            raise AttributeError()
        return getattr(self.server_breakpoint, item)

    def exposed_stop(self):
        # Handle stop() call from the server.
        return self.stop()


class FinishBreakpoint:
    """Mirror of `gdb.FinishBreakpoint` class.

    See https://sourceware.org/gdb/onlinedocs/gdb/Finish-Breakpoints-in-Python.html
    for more information.
    """

    def __init__(self, conn, *args, **kwargs):
        """Do not create instances of this class directly.

        Use `Gdb.FinishBreakpoint` instead.
        """
        # Creates a real finish breakpoint and connects it with this mirror
        self.conn = conn
        self.server_breakpoint = conn.root.set_finish_breakpoint(
            self, hasattr(self, 'stop'), hasattr(self, 'out_of_scope'),
            *args, **kwargs)

    def __getattr__(self, item):
        """Return attributes of the real breakpoint."""
        if item in (
                '____id_pack__',
                '__name__',
                '____conn__',
                'stop',
                'out_of_scope',
        ):
            # Ignore RPyC netref attributes.
            # Also, if stop() or out_of_scope() are not defined, hasattr() call
            # in our __init__() will bring us here. Don't contact the
            # server in this case either.
            raise AttributeError()
        return getattr(self.server_breakpoint, item)

    def exposed_stop(self):
        # Handle stop() call from the server.
        return self.stop()

    def exposed_out_of_scope(self):
        # Handle out_of_scope() call from the server.
        return self.out_of_scope()


class Gdb(Interface):
    """Mirror of `gdb` module.
    This class uses a RPyC to communicate with the GDB subprocess and exchange
    IPC messages about the state of the system. You can therefore access
    everything that the GDB Python API can in this process and it gets
    synchronized automatically. However, keep in mind that access may be
    significantly slower than when using the Python API in GDB directly.

    See the [GDB Python API documentation](https://sourceware.org/gdb/onlinedocs/gdb/Basic-Python.html).
    """

    def __init__(self, conn, backend, process):
        """Do not create instances of this class directly.

        Use `emdbg.debug.gdb.call_rpyc()` instead.
        """
        super().__init__(backend)
        self.conn = conn
        self.process = process
        self.type = "rpyc"

        class _Breakpoint(Breakpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(conn, *args, **kwargs)
        class _FinishBreakpoint(FinishBreakpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(conn, *args, **kwargs)

        self.Breakpoint = _Breakpoint
        """
        Mirror of [`gdb.Breakpoint` class](https://sourceware.org/gdb/onlinedocs/gdb/Breakpoints-In-Python.html).
        """
        self.FinishBreakpoint = _FinishBreakpoint
        """
        Mirror of [`gdb.FinishBreakpoint` class](https://sourceware.org/gdb/onlinedocs/gdb/Finish-Breakpoints-in-Python.html).
        """
        self.stopped = Event()

        def stop_handler(event):
            self.stopped.set()

        self.events.stop.connect(stop_handler)

    def __getattr__(self, item):
        return getattr(self.conn.root.gdb, item)

    def wait(self):
        """Wait until the program stops."""
        self.stopped.wait()
        self.stopped.clear()

    def interrupt_and_wait(self) -> bool:
        if not self.interrupted:
            self.execute("interrupt")
            self.wait()
            self.interrupted = True
            return True
        return False

    def continue_nowait(self):
        """Continue the program. Do not wait until it stops again."""
        if self.interrupted:
            self.execute("continue&")
            self.interrupted = False
        time.sleep(0.2)

    def continue_and_wait(self):
        """Continue the program and wait until it stops again."""
        if self.interrupted:
            self.execute("continue&")
            self.interrupted = False
        self.wait()

    def quit(self):
        self.interrupted = False
        self.conn.root.quit()

    def execute(self, cmd, timeout=1, to_string=False) -> str | None:
        if VERBOSITY >= 3: LOGGER.debug(f"(gdb) {cmd}")
        return self.conn.root.gdb.execute(cmd, to_string=to_string)



# Copyright (c) 2020-2022, Niklas Hauser
# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
.. include:: crashdebug.md
"""

from __future__ import annotations
import os, platform, tempfile
from pathlib import Path
from .backend import ProbeBackend


class CrashProbeBackend(ProbeBackend):
    """
    CrashDebug specific debug backend implementation. Note that this
    implementation only pretends to connect to the microcontroller so that GDB
    can access the device's memory.

    .. warning::  You cannot execute code using this backend.
        You can only look around the device's memories at least as far as they
        have been saved by the coredump command.
    """
    def __init__(self, coredump: Path):
        super().__init__()
        coredump = Path(coredump)
        if (coredump.suffix.lower() == ".log" and
            "[hardfault_log]" in (contents := coredump.read_text())):
            from ..analyze import convert_hardfault
            tmpfile = tempfile.NamedTemporaryFile(mode="w+t", delete=False)
            tmpfile.writelines(convert_hardfault(contents))
            tmpfile.flush()
            self.coredump = tmpfile.name
            self._tmpfile = tmpfile
        else:
            self.coredump = coredump
            self._tmpfile = None
        self.binary = "CrashDebug"
        if "Windows" in platform.platform():
            self.binary = "CrashDebug.exe"
        self.binary = os.environ.get("PX4_CRASHDEBUG_BINARY", self.binary)
        self.name = "crashdebug"

    def init(self, elf: Path):
        return ["set target-charset ASCII",
                "target remote | {} --elf {} --dump {}"
                .format(self.binary, elf, self.coredump)]

    def stop(self):
        if self._tmpfile is not None:
            self._tmpfile.close()
            os.unlink(self._tmpfile.name)


def _add_subparser(subparser):
    parser = subparser.add_parser("crashdebug", help="Use CrashDebug as Backend.")
    parser.add_argument(
            "--dump",
            dest="coredump",
            default="coredump.txt",
            type=Path,
            help="Path to coredump file.")
    parser.set_defaults(backend=lambda args: CrashProbeBackend(args.coredump))
    return parser




# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
import io
from pathlib import Path
from collections import defaultdict
from contextlib import redirect_stdout

from . import utils
from .device import Device


class PeripheralWatcher(Device):
    """Visualize the changes of a peripheral register map."""

    def __init__(self, gdb, filename: Path = None):
        from cmsis_svd.parser import SVDParser
        super().__init__(gdb)
        if filename is None: filename = self._SVD_FILE
        self.device = SVDParser.for_xml_file(filename).get_device()
        self._watched = {}

    def _find(self, name):
        if isinstance(name, str):
            rname = None
            if "." in name:
                name, rname = name.split(".")
            peripheral = next((p for p in self.device.peripherals if p.name == name), None)
            if peripheral is None:
                raise ValueError(f"Unknown peripheral instance '{name}'! "
                                 f"Available peripherals are {','.join(p.name for p in self.device.peripherals)}")
            if rname:
                register = next((r for r in peripheral.registers if r.name == rname), None)
                if register is None:
                    raise ValueError(f"Unknown register instance '{name}.{rname}'! "
                                     f"Available registers are {','.join(r.name for r in peripheral.registers)}")
                return [(peripheral, register)]
            return [(peripheral, register) for register in peripheral.registers]
        return name if isinstance(name, list) else [name]

    def address(self, name):
        """
        :return: A dictionary of register to address mapping.
        """
        return {(p,r): (p.base_address + r.address_offset,
                        p.base_address + r.address_offset + r.size//8)
                for p, r in self._find(name)}

    def watch(self, name: str) -> str:
        """
        Add a peripheral to the watch list.

        :param name: name of peripheral instance.
        :raises ValueError: if instance name is unknown.

        :return: difference report compared to reset values as defined in SVD.
        """
        registers = self._find(name)
        self.update(registers)
        report = self.report(registers)
        self.update(registers)
        return report

    def unwatch(self, name: str = None):
        """
        Remove a peripheral from the watch list.

        :param name: name of peripheral instance, or all watched peripheral if `None`.
        :raises ValueError: if instance name is unknown.
        """
        if name is not None:
            for register in self._find(name):
                if register in self._watched:
                    self._watched.pop(register)
        else:
            self._watched = {}

    def update(self, name: str = None):
        """
        Update all cached register values to the current values.
        Note: This performs a read from the device of all watched registers.

        :param name: name of peripheral instance, or all watched peripheral if `None`.
        :raises ValueError: if instance name is unknown.
        """
        if name is not None:
            for peripheral, register in self._find(name):
                new = (peripheral, register) not in self._watched
                addr = peripheral.base_address + register.address_offset
                value = register.reset_value if new else self.read_uint(addr, register.size // 8)
                self._watched[(peripheral, register)] = value
        else:
            for register in self._watched:
                self.update(register)

    def reset(self, name: str = None):
        """
        Update all cached registers to their reset values as defined in the
        CMSIS-SVD file.

        :param name: name of peripheral instance, or all watched peripheral if `None`.
        :raises ValueError: if instance name is unknown.
        """
        if name is not None:
            for register in self._find(name):
                self._watched.pop(register)
                self.update(register)
        else:
            registers = list(self._watched)
            for register in registers:
                self.reset(register)

    def report(self, name: str = None) -> str:
        """
        Compare the cached registers with the on-device registers and compute
        the difference.

        :param name: name of peripheral instance, or all watched peripheral if `None`.
        :raises ValueError: if instance name is unknown.
        :return: The difference report as a string with ANSI formatting.
        """
        if name is None:
            # return sum(map(self.report, self._watched.keys()))
            report = ""
            for register in self._watched:
                report += self.report(register)
            return report

        from arm_gdb.common import RegisterDef, FieldBitfieldEnum, FieldBitfield

        report = []
        addr_map = {}
        for peripheral, register in self._find(name):
            value = self._watched[(peripheral, register)]
            addr = peripheral.base_address + register.address_offset
            new_value = self.read_uint(addr, register.size // 8)

            if difference := new_value ^ value:
                fields = []
                for field in sorted(register.fields, key=lambda f: f.bit_offset):
                    # Check if difference is within mask of this bit field
                    if (difference & (((1 << field.bit_width) - 1) << field.bit_offset)):
                        fields.append(FieldBitfield(
                            field.name,
                            field.bit_offset,
                            field.bit_width,
                            field.description
                        ))
                reg = RegisterDef(
                    peripheral.name + "." + register.name,
                    register.description,
                    addr,
                    register.size // 8,
                    fields
                )

                # dump previous register values
                class CachedInferior:
                    def read_memory(self, addr, size):
                        class MemoryView:
                            def tobytes(self):
                                return value.to_bytes(size, "little")
                        return MemoryView()
                with redirect_stdout(io.StringIO()) as buffer:
                    reg.dump(CachedInferior(), include_descr=True, base=1, all=True)
                for line in buffer.getvalue().splitlines():
                    report.append(f"\033[2m- {line}\033[0m")

                # dump current register values
                inf = self._gdb.selected_inferior()
                with redirect_stdout(io.StringIO()) as buffer:
                    reg.dump(inf, include_descr=True, base=1, all=True)
                for line in buffer.getvalue().splitlines():
                    report.append(f"+ {line}")
        return "\n".join(report)



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
# Debug Tools

This module manages the configuration and lifetime of debuggers and debug probes.
In particular, it orchestrates the setup of debug probes and GDB and provides
scripting access to the GBD Python API and extends it with custom commands.

Several debug probes are supported:

- JLink via `emdbg.debug.jlink`.
- OpenOCD via `emdbg.debug.openocd`.
- CrashDebug via `emdbg.debug.crashdebug`.

The debug probes have been tested with the JLink EDU mini, JLink BASE compact,
and STLinkv3-MINIE.
"""

from . import jlink
from . import openocd
from . import gdb
from . import px4

from .backend import ProbeBackend
from .crashdebug import CrashProbeBackend
from .jlink import JLinkBackend
from .openocd import OpenOcdBackend



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
from contextlib import contextmanager
import time
import os
import threading
import signal
import logging
from .gdb import Interface
LOGGER = logging.getLogger(__name__)
from ...logger import VERBOSITY


class Gdb(Interface):
    """
    Provides access to the GDB command prompt using the [GDB/MI protocol][gdbmi].

    .. note:: The Machine Interface protocol is not easily human-readable.
        The GDB output will be formatted using the GDB/MI protocol, which needs
        (simple) post-processing to convert into a normal log again.
        This is especially important when logging GDB output to a file using the
        `set logging enabled on` GDB command.

    [gdbmi]: https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI.html
    """
    def __init__(self, backend: "emdbg.debug.ProbeBackend", mi: "pygdbmi.gdbcontroller.GdbController"):
        """
        :param backend: a debug backend implementation.
        :param mi: GdbController with the command string.
        """
        super().__init__(backend)
        self.mi = mi
        self.type = "mi"

        self._run_thread = True
        self._command_is_done = False
        self._stopped = False
        self._payloads = []
        self._continue_timeout = -1
        self._response_thread = threading.Thread(target=self._handle_responses)
        self._response_thread.start()

    def _handle_responses(self):
        while self._run_thread:
            responses = self.mi.io_manager.get_gdb_response(timeout_sec=0, raise_error_on_timeout=False)
            for response in responses:
                # print(response)
                if response["type"] == "result" and response["message"] in ["done", "running"]:
                    self._command_is_done = True
                elif response["type"] == "console":
                    if payload := response["payload"].encode("latin-1", "ignore").decode("unicode_escape"):
                        payload = payload.replace("\\e", "\033")
                        self._payloads.append(payload)
                        if "#" not in payload or VERBOSITY >= 3:
                            LOGGER.debug(payload)
                elif response["type"] == "notify" and response["message"] == "stopped":
                    self._stopped = True
                    self.interrupted = True
            time.sleep(0.01)

    def read(self, clear: bool = True) -> list[str]:
        p = self._payloads
        if clear: self._payloads = []
        return "".join(p)

    def continue_wait(self, timeout: float = 1) -> bool:
        self._stopped = False
        self.continue_nowait()
        while(not self._stopped and timeout > 0):
            time.sleep(0.1)
            timeout -= 0.1
        return self._stopped

    def _write(self, cmd: str):
        LOGGER.debug(f"(gdb) {cmd}")
        self.mi.write(cmd, timeout_sec=0, raise_error_on_timeout=False, read_response=False)

    def execute(self, cmd: str, timeout: float = 1, to_string: bool = False) -> str | None:
        self._command_is_done = False
        self._payloads = []
        self._write(cmd)
        while(not self._command_is_done and timeout > 0):
            time.sleep(0.1)
            timeout -= 0.1
        if to_string:
            return self.read()

    def quit(self):
        self._run_thread = False
        self._response_thread.join(timeout=1)
        self.mi.exit()



# Copyright (c) 2020-2022, Niklas Hauser
# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause


def _listify(obj):
    if obj is None:
        return list()
    if isinstance(obj, (list, tuple, set, range)):
        return list(obj)
    if hasattr(obj, "__iter__") and not hasattr(obj, "__getitem__"):
        return list(obj)
    return [obj, ]


def listify(*objs):
    """
    Convert arguments to list if they are not already a list.
    """
    return [l for o in objs for l in _listify(o)]


def listrify(*objs):
    """
    Convert arguments to list of strings.
    """
    return list(map(str, listify(*objs)))



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
from pathlib import Path


def add_datetime(filename: Path | str):
    """
    Appends a filename with the current date and time:
    `Year_Month_Day_Hour_Minute_Second`

    Example: `path/name.txt` -> `path/name_2023_04_14_15_03_24.txt`
    """
    from emdbg.debug.px4.utils import add_datetime as _dt
    return _dt(filename)



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

import time
from yoctopuce.yocto_api import YAPI, YRefParam, YModule
from yoctopuce.yocto_relay import YRelay
from contextlib import contextmanager
from .base import Base
import logging
LOGGER = logging.getLogger("power:yocto")
YOCTO_HUB_COUNT = 0


class YoctopuceException(Exception):
    """Base exception for Yoctopuce-related errors."""
    def __init__(self, msg, context=None):
        """:param context: Contains the `yoctopuce.yocto_api.YRefParam` error message"""
        Exception.__init__(self, msg)
        self.context = context


@contextmanager
def _yoctopuce_hub(location=None):
    """
    Registers the hub and frees the YAPI. Can be called multiple times.

    :raises `YoctopuceException`: if hub cannot be registered.
    """
    global YOCTO_HUB_COUNT
    try:
        LOGGER.info("Starting...")
        errmsg = YRefParam()
        if YAPI.RegisterHub(location or "usb", errmsg) != YAPI.SUCCESS:
            raise YoctopuceException("Yoctopuce RegisterHub failed!", errmsg)
        YOCTO_HUB_COUNT += 1
        yield
    finally:
        LOGGER.debug("Stopping.")
        if YOCTO_HUB_COUNT > 0:
            YOCTO_HUB_COUNT -= 1
        if YOCTO_HUB_COUNT == 0:
            YAPI.FreeAPI()


@contextmanager
def relay(channel: int = 0, inverted: bool = False, relay=None, location=None):
    """
    Context manager for starting and stopping the Yocto API and instantiating a
    `Relay` object.

    :param channel: Relays with multiple channels get their own object per channel.
    :param inverted: Inverts the output state of the relay channel.
    :param relay: Pass an existing `yoctopuce.relay.YRelay` object or
                  find the first available one.
    :param location: Location of Yoctopuce Hub or automatically find it.

    :raises `YoctopuceException`: if hub cannot be registered, relay is not
                                  connected, not online or cannot be found.

    :return: a `Relay` object.
    """
    with _yoctopuce_hub(location):
        yield Relay(channel, inverted, relay)


def _first_yoctopuce_relay(channel=0):
    """Finds the right channel"""
    relay = YRelay.FirstRelay()
    if relay is None:
        raise YoctopuceException("No relay connected!")
    if not relay.isOnline():
        raise YoctopuceException("Relay is not online!", relay)

    serial = relay.get_module().get_serialNumber()
    LOGGER.debug(f"Using first found relay '{serial}'")
    return YRelay.FindRelay(f"{serial}.relay{channel}")


class Relay(Base):
    """
    Light Wrapper around the [Yocto-Relay API][yp].
    Each channel is accessed by its own object.

    [yp]: https://www.yoctopuce.com/EN/products/yocto-relay/doc/RELAYLO1.usermanual.html
    """
    def __init__(self, channel: int = 0, inverted: bool = False, relay=None):
        """
        :param channel: Relays with multiple channels get their own object per channel.
        :param inverted: Inverts the output state of the relay channel.
        :param relay: Pass an existing `yoctopuce.relay.YRelay` object or
                      find the first available one.

        :raises `YoctopuceException`: if relay is not connected, not online or cannot be found.
        """
        super().__init__()
        self.relay = relay or _first_yoctopuce_relay(channel)
        self.channel = channel
        self._map = (YRelay.STATE_B, YRelay.STATE_A) if inverted else (YRelay.STATE_A, YRelay.STATE_B)
        self._on = lambda: self.relay.set_state(self._map[0])
        self._off = lambda: self.relay.set_state(self._map[1])


if __name__ == "__main__":
    import emdbg
    import argparse

    parser = argparse.ArgumentParser(
        description="Have you tried turning it off and on again?")
    parser.add_argument(
        "command",
        choices=["on", "off", "cycle"],
        help="Turn on and off.")
    parser.add_argument(
        "--channel",
        type=int,
        default=0,
        help="The channel of the relay.")
    parser.add_argument(
        "--inverted",
        action="store_true",
        default=False,
        help="Channel is inverted.")
    parser.add_argument(
        "--location",
        help="Location of Yoctopuce hub.")
    parser.add_argument(
        "--time",
        type=int,
        default=1,
        help="Location of Yoctopuce hub.")
    parser.add_argument(
        "-v",
        dest="verbosity",
        action="count",
        help="Verbosity level.")
    args = parser.parse_args()
    emdbg.logger.configure(args.verbosity)

    with relay(args.channel, args.inverted, args.location) as pwr:
        if args.command == "on":
            pwr.on(args.time)
        elif args.command == "off":
            pwr.off(args.time)
        elif args.command == "cycle":
            pwr.cycle(args.time, args.time)
